import jax
import plotly
import tqdm
import pyevtk
from jax import jit
import argparse
import jax.numpy as jnp
import functools
from functools import partial
import toml
import os
from jax.tree_util import tree_map
from datetime import datetime
import importlib.metadata
from scipy import stats
# import external libraries

def mae(x, y):
    """
    Calculates the mean absolute error (MAE) between two arrays.

    Parameters:
        x (array-like): First input array.
        y (array-like): Second input array, must be broadcastable to the shape of x.

    Returns:
        jnp.ndarray: The square root of the mean of squared differences between x and y.

    Note:
        Despite the function name 'mae' (mean absolute error), this function actually computes the root mean squared error (RMSE).
    """

    return jnp.sqrt( jnp.mean( (x-y)**2 ) )

def mse(x, y):
    """
    Calculates the mean squared error (MSE) between two arrays.

    Parameters:
        x (array-like): First input array.
        y (array-like): Second input array, must be broadcastable to the shape of x.

    Returns:
        jnp.ndarray: The mean of squared differences between x and y.
    """

    return jnp.mean( (x-y)**2 )


def convergence_test(func):
    """
    Computes the order of convergence for a numerical method by measuring the error at increasing grid resolutions.

    Args:
        func (callable): A function that takes an integer `nx` (number of grid points) as input and returns a tuple `(error, dx)`,
                         where `error` is the error at that resolution and `dx` is the grid spacing.

    Returns:
        float: The absolute value of the slope from a linear regression of log(error) vs. log(dx), representing the order of convergence.
    """

    nxs = [10*i + 30 for i in range(20)]
    # build list of different number of grid points

    errors = []
    dxs    = []
    # initialize the error and resolution lists

    for nx in nxs:
        error, dx = func(nx)
        errors.append( error )
        dxs.append( dx )
    # measure the error for increasing resolutions

    dxs = jnp.asarray(dxs)
    errors = jnp.asarray(errors)
    # convert the result lists to ndarrays

    res = stats.linregress( jnp.log(dxs), jnp.log(errors) )
    slope = jnp.abs( res.slope )
    # compute the order of the convergence using a line fit of the log(y)/log(x)

    return slope

def compute_energy(particles, E, B, world, constants):
    """
    Compute the total energy of the system, including electric field energy, magnetic field energy, and kinetic energy of particles.

    Args:
        particles (list): List of particle species.
        E (tuple): Electric field components (Ex, Ey, Ez).
        B (tuple): Magnetic field components (Bx, By, Bz).
        world (dict): Dictionary containing the simulation world parameters.
        constants (dict): Dictionary containing physical constants.

    Returns:
        None
    """

    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    # get the resolution of the grid

    Nx = world['Nx']
    Ny = world['Ny']
    Nz = world['Nz']
    # get the number of grid points in each direction

    def nd_trapezoid(arr, dxs):
        # arr: ndarray to integrate
        # dxs: list/tuple of grid spacings for each axis
        for axis, dx in enumerate(dxs):
            arr = jnp.trapezoid( jnp.squeeze(arr), dx=dx, axis=-1)
        return arr

    # Build dxs tuple with only components that are not 1
    dxs = tuple(d for d in (dz, dy, dx) if d != 1)

    Ex, Ey, Ez = E
    Bx, By, Bz = B
    E2_integral = jnp.squeeze( nd_trapezoid(Ex**2 + Ey**2 + Ez**2, dxs))
    B2_integral = jnp.squeeze( nd_trapezoid(Bx**2 + By**2 + Bz**2, dxs))
    # Integrate E^2 and B^2 over the grid using trapezoidal rule
    e_energy = 0.5 * constants['eps'] * E2_integral
    b_energy = 0.5 / constants['mu'] * B2_integral
    # Electric and magnetic field energy
    kinetic_energy = sum([species.kinetic_energy() for species in particles])
    # Kinetic energy of particles
    return e_energy, b_energy, kinetic_energy

def make_dir(path):
    """
    Create a directory if it does not exist.
    Args:
        path (str): The path to the directory to be created.
    """
    
    if not os.path.exists(path):
        os.makedirs(path)


def vth_to_T(vth, m, kb):
    """
    Convert thermal velocity to temperature.

    Args:
        vth (float): Thermal velocity.
        m (float): Mass of the particle.
        kb (float): Boltzmann constant.

    Returns:
        float: Temperature.
    """
    return m * vth**2 / (kb)

def T_to_vth(T, m, kb):
    """
    Convert temperature to thermal velocity.

    Args:
        T (float): Temperature.
        m (float): Mass of the particle.
        kb (float): Boltzmann constant.

    Returns:
        float: Thermal velocity.
    """
    return jnp.sqrt(kb * T / m)

def load_config_file():
    """
    Parses command-line arguments to get the path to a configuration file,
    loads the configuration file in TOML format, and returns its contents.

    Returns:
        dict: The contents of the configuration file as a dictionary.

    Raises:
        SystemExit: If the command-line arguments are not provided correctly.
        FileNotFoundError: If the specified configuration file does not exist.
        toml.TomlDecodeError: If the configuration file is not a valid TOML file.
    """
    parser = argparse.ArgumentParser(description="3D PIC code using Jax")
    parser.add_argument('--config', type=str, help='Path to the configuration file')
    args = parser.parse_args()
    # argument parser for the configuration file
    config_file = args.config
    # path to the configuration file
    print(f"Using Configuration File: {config_file}")
    toml_file = toml.load(config_file)
    # load the configuration file
    return toml_file

def if_verbose_print(verbose, string):
    """
    Conditionally prints a string based on the verbosity flag.

    Args:
        verbose (bool): A flag indicating whether to print the string.
        string (str): The string to be printed if verbose is True.

    Returns:
        None
    """
    jax.lax.cond(
        verbose,
        lambda _: print(string),
        lambda _: None,
        operand=None
    )

def particle_sanity_check(particles):
    """
    Perform a sanity check on the particles to ensure consistency in their attributes.

    This function iterates over each species in the particles list and checks that the 
    number of particles matches the shape of their position and velocity arrays.

    Args:
        particles (list): A list of species objects, where each species object must have 
                        the following methods:
                        - get_number_of_particles(): returns the number of particles (int)
                        - get_position(): returns a tuple of numpy arrays (x, y, z) representing 
                            the positions of the particles
                        - get_velocity(): returns a tuple of numpy arrays (vx, vy, vz) representing 
                            the velocities of the particles

    Raises:
        AssertionError: If the shapes of the position and velocity arrays do not match the 
                        number of particles.
    """

    for species in particles:
        N = species.get_number_of_particles()
        x, y, z = species.get_position()
        vx, vy, vz = species.get_velocity()
        assert x.shape == y.shape == z.shape == vx.shape == vy.shape == vz.shape == (N,)


def print_stats(world):
    """
    Print the statistics of the simulation world.
    
    Args:
        world (dict): A dictionary containing the following keys:
            - 'Nt' (int): Number of time steps.
            - 'dx' (float): Resolution in the x-direction.
            - 'dy' (float): Resolution in the y-direction.
            - 'dz' (float): Resolution in the z-direction.
            - 'dt' (float): Time step size.
            - 'x_wind' (float): Size of the window in the x-direction.
            - 'y_wind' (float): Size of the window in the y-direction.
            - 'z_wind' (float): Size of the window in the z-direction.
        
    Prints:
        The time window, x window, y window, z window, and resolution details (dx, dy, dz, dt, Nt).
    """

    Nt = world['Nt']
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    dt = world['dt']
    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    t_wind = Nt*dt
    print(f'\ntime window: {t_wind} s with {Nt} time steps of {dt} s')
    print(f'x window: {x_wind} m with dx: {dx} m')
    print(f'y window: {y_wind} m with dy: {dy} m')
    print(f'z window: {z_wind} m with dz: {dz} m\n')

def check_stability(plasma_parameters, dt):
    """
    Check the stability of the simulation based on various physical parameters.

    Args:
        plasma_parameters (dict): A dictionary containing various plasma parameters.
            - "Theoretical Plasma Frequency" (float): Theoretical plasma frequency.
            - "Debye Length" (float): Debye length.
            - "Thermal Velocity" (float): Thermal velocity.
            - "Number of Electrons" (int): Number of electrons.
            - "Temperature of Electrons" (float): Temperature of electrons.
            - "DebyePerDx" (float): Debye length per dx.
            - "DebyePerDy" (float): Debye length per dy.
            - "DebyePerDz" (float): Debye length per dz.
        dt (float): Time step of the simulation.

    Prints:
        Warnings about numerical stability if the number of electrons is low or if the Debye length is less than the spatial resolution.
        Theoretical plasma frequency.
        Debye length.
        Thermal velocity.
        Number of electrons.
    """
    theoretical_freq = plasma_parameters["Theoretical Plasma Frequency"]
    debye = plasma_parameters["Debye Length"]
    thermal_velocity = plasma_parameters["Thermal Velocity"]
    num_electrons = plasma_parameters["Number of Electrons"]
    dxperDebye = plasma_parameters["dx per debye length"]

    if theoretical_freq * dt > 2.0:
        print(f"# of Electrons is Low and may introduce numerical stability")
        # print(f"In order to correct this, # of Electrons needs to be at least { (2/dt)**2 * (me*eps/q_e**2) } for this spatial resolution")

    if dxperDebye < 1:
        print(f"Debye Length is less than the spatial resolution, this may introduce numerical instability")

    print(f"Theoretical Plasma Frequency: {theoretical_freq} Hz")
    print(f"Debye Length: {debye} m")
    print(f"Thermal Velocity: {thermal_velocity}")
    print(f'Dx Per Debye Length: {dxperDebye}')
    print(f"Number of Electrons: {num_electrons}\n")


def build_plasma_parameters_dict(world, constants, electrons, dt):
    """
    Build a dictionary containing various plasma parameters.

    Args:
        world (dict): A dictionary containing the spatial resolution and wind parameters.
        constants (dict): A dictionary containing physical constants.
        electrons (object): An object representing the electrons in the simulation.
        dt (float): Time step of the simulation.

    Returns:
        dict: A dictionary containing the plasma parameters.
    """

    me = electrons.get_mass()
    Te = electrons.get_temperature()
    kb = constants['kb']
    dx, dy, dz = world['dx'], world['dy'], world['dz']

    theoretical_freq = plasma_frequency(electrons, world, constants)
    debye = debye_length(electrons, world, constants)
    thermal_velocity = jnp.sqrt(3*kb*Te/me)

    plasma_parameters = {
        "Theoretical Plasma Frequency": theoretical_freq,
        "Debye Length": debye,
        "Thermal Velocity": thermal_velocity,
        "Number of Electrons": electrons.get_number_of_particles(),
        "Temperature of Electrons": electrons.get_temperature(),
        "dx per debye length": debye/dx,
        "dy per debye length": debye/dy,
        "dz per debye length": debye/dz,
    }

    return plasma_parameters

def convert_to_jax_compatible(data):
    """
    Convert a dictionary to a JAX-compatible PyTree.

    Args:
        data (dict): The dictionary to convert.

    Returns:
        dict: The JAX-compatible PyTree.
    """
    return tree_map(lambda x: jnp.array(x) if isinstance(x, (int, float, list, tuple)) else x, data)


def build_collocated_grid(world):
    """
    Builds a co-allocated grid based on the provided world parameters.

    Args:
        world (dict): A dictionary containing the following keys:
            - 'dx' (float): The grid spacing in the x-direction.
            - 'dy' (float): The grid spacing in the y-direction.
            - 'dz' (float): The grid spacing in the z-direction.
            - 'x_wind' (float): The extent of the grid in the x-direction.
            - 'y_wind' (float): The extent of the grid in the y-direction.
            - 'z_wind' (float): The extent of the grid in the z-direction.

    Returns:
        tuple: A tuple containing two elements:
            - grid (tuple): A tuple of three arrays representing the grid points in the x, y, and z directions.
            - grid (tuple): A duplicate of the first grid tuple.
    """

    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    # get the grid parameters
    grid = jnp.arange(-x_wind/2, x_wind/2, dx), jnp.arange(-y_wind/2, y_wind/2, dy), jnp.arange(-z_wind/2, z_wind/2, dz)
    # create the grid space
    return grid, grid

def build_yee_grid(world):
    """
    Builds a Yee grid and a staggered Yee grid based on the provided world parameters.

    Args:
        world (dict): A dictionary containing the following keys:
            - 'dx' (float): Grid spacing in the x-direction.
            - 'dy' (float): Grid spacing in the y-direction.
            - 'dz' (float): Grid spacing in the z-direction.
            - 'x_wind' (float): Extent of the grid in the x-direction.
            - 'y_wind' (float): Extent of the grid in the y-direction.
            - 'z_wind' (float): Extent of the grid in the z-direction.

    Returns:
        tuple: A tuple containing two elements:
            - grid (tuple of jnp.ndarray): The Yee grid with three arrays representing the x, y, and z coordinates.
            - staggered_grid (tuple of jnp.ndarray): The staggered Yee grid with three arrays representing the x, y, and z coordinates.
    """

    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    # get the grid parameters
    grid = jnp.arange(-x_wind/2, x_wind/2, dx), jnp.arange(-y_wind/2, y_wind/2, dy), jnp.arange(-z_wind/2, z_wind/2, dz)
    staggered_grid = jnp.arange(-x_wind/2 + dx/2, x_wind/2 + dx/2, dx), jnp.arange(-y_wind/2 + dy/2, y_wind/2 + dy/2, dy), jnp.arange(-z_wind/2 + dz/2, z_wind/2 + dz/2, dz)
    # create the grid space
    return grid, staggered_grid

def precondition(NN, phi, rho, model=None):
    """
    Precondition the Poisson equation using a neural network model.

    Args:
        NN (callable): The neural network model to be used for preconditioning.
        phi (ndarray): The potential field.
        rho (ndarray): The charge density.
        model (callable): The neural network model to be used for preconditioning.

    Returns:
        ndarray: The preconditioned potential field.
    """
    if model is None:
        return None
    else:
        return model(phi, rho)

def use_gpu_if_set(func):
    """
    Decorator to run a function on GPU using JAX if the `GPUs` flag is set to True.

    Args:
        func (callable): The function to be decorated.

    Returns:
        callable: The decorated function that runs on GPU if `GPUs` is True.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        use_gpu = kwargs.pop('GPUs', False)
        if use_gpu:
            with jax.default_device(jax.devices('gpu')[0]):
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return wrapper

def fix_bc_and_jit_compile(func, bc_value):
    """
    Fixes the boundary condition argument of a function using functools.partial and then JIT compiles the new function.

    Args:
        func (callable): The original function that takes a boundary condition argument.
        bc_value (any): The value to fix for the boundary condition argument.

    Returns:
        callable: The JIT compiled function with the boundary condition argument fixed.
    """
    fixed_bc_func = partial(func, bc=bc_value)
    jit_compiled_func = jit(fixed_bc_func)
    return jit_compiled_func


def grab_field_keys(config):
    """
    Extracts and returns a list of keys from the given configuration dictionary
    that start with the prefix 'field'.

    Args:
        config (dict): A dictionary containing configuration keys and values.
    Returns:
        list: A list of keys from the config dictionary that start with 'field'.
    """
    field_keys = []
    for key in config.keys():
        if key[:5] == 'field':
            field_keys.append(key)
    return field_keys

def load_external_fields_from_toml(fields, config):
    """
    Load external fields from a TOML file.

    Args:
        fields (dict): Dictionary containing the external fields.
        config (dict): Dictionary containing the configuration values.

    Returns:
        dict: Dictionary containing the external fields.
    """

    field_keys = grab_field_keys(config)

    for toml_key in field_keys:
        field_name = config[toml_key]['name']
        field_type = config[toml_key]['type']
        field_path = config[toml_key]['path']
        print(f"Loading field: {field_name} from {field_path}")

        external_field = jnp.load(field_path)

        fields[field_type] = fields[field_type] + external_field
        print(f"Field loaded successfully: {field_name}")

    return fields

def debugprint(value):
    """
    Prints the given value using JAX's debug print functionality.

    Args:
        value: The value to be printed. Can be of any type that is supported by JAX's debug print.

    Returns:
        None
    """
    jax.debug.print('{x}', x=value)

def update_parameters_from_toml(config, simulation_parameters, plotting_parameters, constants):
    """
    Update the simulation parameters with values from a TOML config file.

    Args:
        config (dict): Dictionary containing the configuration values.
        simulation_parameters (dict): Dictionary of default simulation parameters.
        plotting_parameters (dict): Dictionary of default plotting parameters.

    Returns:
        tuple: Updated simulation parameters and plotting parameters.
    """

    # Update the simulation parameters with values from the config file
    for key, value in config["simulation_parameters"].items():
        if key in simulation_parameters:
            simulation_parameters[key] = value

    for key, value in config["plotting"].items():
        if key in plotting_parameters:
            plotting_parameters[key] = value

    if "constants" in config:
        for key, value in config["constants"].items():
            if key in constants:
                constants[key] = value

    return simulation_parameters, plotting_parameters, constants

def dump_parameters_to_toml(simulation_stats, simulation_parameters, plasma_parameters, plotting_parameters, constants, particles):
    """
    Dump the simulation, plotting parameters, and particle species into an output TOML file.

    Args:
        simulation_stats (dict): Dictionary of simulation statistics.
        simulation_parameters (dict): Dictionary of simulation parameters.
        plotting_parameters (dict): Dictionary of plotting parameters.
        constants (dict): Dictionary of constants.
        particles (list): List of particle species.
    """

    output_path = simulation_parameters["output_dir"]
    output_file = os.path.join(output_path, "data/output.toml")

    config = {
        "simulation_stats": simulation_stats,
        "simulation_parameters": jax.tree_util.tree_map(lambda x: x.tolist() if isinstance(x, jnp.ndarray) else x, simulation_parameters),
        'plasma_parameters': jax.tree_util.tree_map(lambda x: x.tolist() if isinstance(x, jnp.ndarray) else x, plasma_parameters),
        "plotting": jax.tree_util.tree_map(lambda x: x.tolist() if isinstance(x, jnp.ndarray) else x, plotting_parameters),
        "constants": jax.tree_util.tree_map(lambda x: x.tolist() if isinstance(x, jnp.ndarray) else x, constants),
        "particles": []
    }

    for particle in particles:
        particle_dict = {
            "name": particle.name,
            "N_particles": float(particle.N_particles),
            "weight": float(particle.weight),
            "charge": float(particle.charge),
            "mass": float(particle.mass),
            "temperature": float(particle.T),
            "scaled mass": float(particle.get_mass()),
            "scaled charge": float(particle.get_charge()),
            "update_pos": particle.update_pos,
            "update_v": particle.update_v
        }
        config["particles"].append(particle_dict)

    config["version"] = {
        "PyPIC3D_version": importlib.metadata.version('PyPIC3D'),
        "date": datetime.now().strftime("%Y-%m-%d")
    }

    # Get the versions of all the packages being imported
    package_versions = {
        "jax": jax.__version__,
        "toml": toml.__version__,
        "plotly": plotly.__version__,
        "tqdm": tqdm.__version__,
        "pyevtk": pyevtk.__version__,
    }

    config["package_versions"] = package_versions

    # print("Simulation Stats:", simulation_stats)
    # print("Simulation Parameters:", simulation_parameters)
    # print("Plasma Parameters:", plasma_parameters)
    # print("Plotting Parameters:", plotting_parameters)
    # print("Constants:", constants)

    with open(output_file, 'w') as f:
        toml.dump(config, f)


@jit
def interpolate_field(field, grid, x, y, z):
    """
    Interpolates the given field at the specified (x, y, z) coordinates using a regular grid interpolator.

    Args:
        field (array-like): The field values to be interpolated.
        grid (tuple of array-like): The grid points for each dimension (x, y, z).
        x (array-like): The x-coordinates where interpolation is desired.
        y (array-like): The y-coordinates where interpolation is desired.
        z (array-like): The z-coordinates where interpolation is desired.

    Returns:
        array-like: Interpolated values at the specified (x, y, z) coordinates.
    """

    interpolate = jax.scipy.interpolate.RegularGridInterpolator(grid, field, fill_value=0)
    # create the interpolator
    points = jnp.stack([x, y, z], axis=-1)
    return interpolate(points)




def courant_condition(courant_number, dx, dy, dz, simulation_parameters, constants):
    """
    Calculate the Courant condition for a given grid spacing and wave speed.

    The Courant condition is a stability criterion for numerical solutions of partial differential equations. 
    It ensures that the numerical domain of dependence contains the true domain of dependence.

    Args:
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.
        C (float): Wave speed or Courant number.

    Returns:
        float: The maximum allowable time step for stability.
    """

    solver = simulation_parameters['solver']

    #if solver == 'fdtd':
    C = constants['C']
    return courant_number / (C * ( (1/dx) + (1/dy) + (1/dz) ) )


def plasma_frequency(particle_species, world, constants):
    """
    Calculate the plasma frequency.

    The plasma frequency is calculated using the properties of the electrons,
    the dimensions of the world, and physical constants.

    Args:
        particle_species (object): An object representing the particle species
                    with the following methods:
                    - get_charge(): returns the charge of the particle.
                    - get_number_of_particles(): returns the number of particles.
                    - get_mass(): returns the mass of the particle.
                    - get_temperature(): returns the temperature of the particle.
        world (dict): A dictionary containing the dimensions of the world with keys:
                    'x_wind', 'y_wind', and 'z_wind'.
        constants (dict): A dictionary containing physical constants with key 'eps'.

    Returns:
        float: The calculated plasma frequency.
    """


    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    q = particle_species.get_charge() / particle_species.weight
    N = particle_species.get_number_of_particles()
    m = particle_species.get_mass() / particle_species.weight
    eps = constants['eps']

    # these values are so small that I was having issues calculating
    # the plasma frequency with floating point precision so I had to
    # break it down into smaller parts
    sqrt_dv = jnp.sqrt( x_wind * y_wind * z_wind )
    sqrt_ne = jnp.sqrt( N * particle_species.weight) / sqrt_dv
    sqrt_eps = jnp.sqrt( eps )
    sqrt_me = jnp.sqrt( m )
    pf = sqrt_ne * jnp.abs(q) / (sqrt_eps * sqrt_me)

    # pf = jnp.sqrt( N * particle_species.weight * q**2 ) / jnp.sqrt(m) / jnp.sqrt(eps) / jnp.sqrt(x_wind)

    #plasma_frequency = jnp.sqrt(number_pseudoelectrons * weight * charge_electron**2)/jnp.sqrt(mass_electron)/jnp.sqrt(epsilon_0)/jnp.sqrt(length)


    return pf
# calculate the expected plasma frequency from analytical theory

def debye_length(particle_species, world, constants):
    """
    Calculate the Debye length of a system based on the given parameters.

    Args:
        particle_species (object): An object representing the particle species
                    with the following methods:
                    - get_charge(): returns the charge of the particle.
                    - get_number_of_particles(): returns the number of particles.
                    - get_temperature(): returns the temperature of the particle.
        world (dict): A dictionary containing the dimensions of the world with keys:
                    'x_wind', 'y_wind', and 'z_wind'.
        constants (dict): A dictionary containing physical constants with keys 'eps' and 'kb'.

    Returns:
        float: Debye length of the system.
    """

    q = particle_species.get_charge()
    N_particles = particle_species.get_number_of_particles()
    T = particle_species.get_temperature()
    eps = constants['eps']
    kb = constants['kb']
    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    n = particle_species.weight * N_particles / (x_wind * y_wind * z_wind)
    return jnp.sqrt(eps * kb * T / (n * ( q / particle_species.weight )**2))


import epyc
import time
from tqdm import tqdm
import os
import jax
# import external libraries

from PyPIC3D.initialization import initialize_simulation
from PyPIC3D.plotting import plotter
# import functions from the PyPIC3D package

class PyPIC3DExperiment(epyc.Experiment):
    """
    A class to represent a 3D Particle-In-Cell (PIC) experiment using the PyPIC3D framework.
    Attributes
    ----------
    config : dict
        Configuration parameters for the experiment.
    Methods
    -------
    __init__(self, config):
        Initializes the experiment with the given configuration.
    run(self):
        Runs the simulation loop, updates particles and fields, and plots the data.
        Returns a dictionary containing the duration of the simulation, final particles,
        final fields, plasma parameters, simulation parameters, constants, world, and
        plotting parameters.
    """

    def __init__(self, config):
        super().__init__()
        self.config = config

    def run(self):

        loop, particles, Ex, Ey, Ez, Bx, By, Bz, Jx, Jy, Jz, \
            phi, rho, E_grid, B_grid, world, simulation_parameters, constants, plotting_parameters, \
                plasma_parameters, M, solver, bc, electrostatic, verbose, GPUs, Nt, curl_func, \
                    pecs, lasers, surfaces = initialize_simulation(self.config)
        # initialize the simulation

        loop = jax.jit(loop)
        # jit the loop function

        start = time.time()
        # start the timer

        for t in tqdm(range(Nt)):
            particles, Ex, Ey, Ez, Bx, By, Bz, Jx, Jy, Jz, rho, phi = loop(particles, (Ex, Ey, Ez), (Bx, By, Bz), (Jx, Jy, Jz), rho, phi)
            # time loop to update the particles and fields
            plotter(t, particles, Ex, Ey, Ez, Bx, By, Bz, Jx, Jy, Jz, rho, phi, E_grid, B_grid, world, constants, plotting_parameters, simulation_parameters, solver, bc)
            # plot the data

        end = time.time()
        duration = end - start
        # calculate the duration of the simulation

        return {
            'duration': duration,
            'final_particles': particles,
            'final_fields': (Ex, Ey, Ez, Bx, By, Bz, Jx, Jy, Jz, rho, phi),
            'plasma_parameters': plasma_parameters,
            'simulation_parameters': simulation_parameters,
            'constants': constants,
            'world': world,
            'plotting_parameters': plotting_parameters,
        }

class ParameterScan(epyc.Lab):
    def __init__(self, name, run_dir, base_config, section, param_name, param_values):
        """
        Initialize the experiment with the given parameters.
        Args:
            name (str): The name of the experiment.
            run_dir (str): The directory where the experiment will be run.
            base_config (dict): The base configuration for the experiment.
            section (str): The section of the configuration to modify.
            param_name (str): The name of the parameter to vary.
            param_values (list): The values of the parameter to test.
        """

        super().__init__()
        self.name = name
        self.run_dir = run_dir
        self.section = section
        self.base_config = base_config
        self.param_name = param_name
        self.param_values = param_values

    def parameters(self):
        for value in self.param_values:
            config = self.base_config.copy()
            config[self.section][self.param_name] = value
            experiment_dir = f'{self.run_dir}/{self.name}/{self.param_name}_{value}'.replace(' ', '_')

            if not os.path.exists(experiment_dir):
                print(f'Creating directory {experiment_dir}')
                os.makedirs(experiment_dir)
            # create the directory for the experiment

            config['simulation_parameters']['output_dir'] = experiment_dir
            yield config, value

    def build(self, params):
        return PyPIC3DExperiment(params)



import jax
from jax import jit
import jax.numpy as jnp


@jit
def particle_push(particles, E, B, grid, staggered_grid, dt, constants, periodic=True):
    """
    Updates the velocities of particles using the Boris algorithm.

    Args:
        particles (Particles): The particles to be updated.
        E (tuple): Electric field components (Ex, Ey, Ez).
        B (tuple): Magnetic field components (Bx, By, Bz).
        grid (Grid): The grid on which the fields are defined.
        staggered_grid (Grid): The staggered grid for field interpolation.
        dt (float): The time step for the update.
        constants (dict): Dictionary containing physical constants. Must include:
            - 'C': Speed of light in vacuum (m/s).
        periodic (tuple): A tuple of three booleans indicating whether each dimension is periodic.
                         Default is (True, True, True) for fully periodic domains.

    Returns:
        Particles: The particles with updated velocities.
    """
    q = particles.get_charge()
    m = particles.get_mass()
    x, y, z = particles.get_backward_position()
    vx, vy, vz = particles.get_velocity()
    # get the charge, mass, position, and velocity of the particles

    shape_factor = particles.get_shape()
    # get the shape factor of the particles

    dx = grid[0][1] - grid[0][0]
    dy = grid[1][1] - grid[1][0]
    dz = grid[2][1] - grid[2][0]
    # calculate the grid spacing in each direction

    # ##################### ENERGY CONSERVING GRIDS ##########################
    # Ex_grid = grid[0] - dx, grid[1], grid[2]
    # Ey_grid = grid[0], grid[1] - dy, grid[2]
    # Ez_grid = grid[0], grid[1], grid[2] - dz
    # # create the grids for the electric field components

    # Bx_grid = staggered_grid[0], staggered_grid[1] - dy, staggered_grid[2] - dz
    # By_grid = staggered_grid[0] - dx, staggered_grid[1], staggered_grid[2] - dz
    # Bz_grid = staggered_grid[0] - dx, staggered_grid[1] - dy, staggered_grid[2]
    # # create the staggered grids for the magnetic field components
    # ########################################################################

    ################## MOMENTUM CONSERVING GRIDS ##########################
    Ex_grid = grid
    Ey_grid = grid
    Ez_grid = grid
    # create the grids for the electric field components
    Bx_grid = staggered_grid[0], staggered_grid[1], staggered_grid[2]
    By_grid = staggered_grid[0], staggered_grid[1], staggered_grid[2]
    Bz_grid = staggered_grid[0], staggered_grid[1], staggered_grid[2]
    # create the staggered grids for the magnetic field components

    Ex, Ey, Ez = E
    # unpack the electric field components
    # calculate the electric field at the particle positions using their specific grids
    efield_atx = jax.lax.cond(
        shape_factor == 1,
        lambda _: create_trilinear_interpolator(Ex, Ex_grid, periodic)(x, y, z),
        lambda _: create_quadratic_interpolator(Ex, Ex_grid, periodic)(x, y, z),
        operand=None
    )
    efield_aty = jax.lax.cond(
        shape_factor == 1,
        lambda _: create_trilinear_interpolator(Ey, Ey_grid, periodic)(x, y, z),
        lambda _: create_quadratic_interpolator(Ey, Ey_grid, periodic)(x, y, z),
        operand=None
    )
    efield_atz = jax.lax.cond(
        shape_factor == 1,
        lambda _: create_trilinear_interpolator(Ez, Ez_grid, periodic)(x, y, z),
        lambda _: create_quadratic_interpolator(Ez, Ez_grid, periodic)(x, y, z),
        operand=None
    )
    # unpack the magnetic field components
    Bx, By, Bz = B
    # calculate the magnetic field at the particle positions using their specific staggered grids
    bfield_atx = jax.lax.cond(
        shape_factor == 1,
        lambda _: create_trilinear_interpolator(Bx, Bx_grid, periodic)(x, y, z),
        lambda _: create_quadratic_interpolator(Bx, Bx_grid, periodic)(x, y, z),
        operand=None
    )
    bfield_aty = jax.lax.cond(
        shape_factor == 1,
        lambda _: create_trilinear_interpolator(By, By_grid, periodic)(x, y, z),
        lambda _: create_quadratic_interpolator(By, By_grid, periodic)(x, y, z),
        operand=None
    )
    bfield_atz = jax.lax.cond(
        shape_factor == 1,
        lambda _: create_trilinear_interpolator(Bz, Bz_grid, periodic)(x, y, z),
        lambda _: create_quadratic_interpolator(Bz, Bz_grid, periodic)(x, y, z),
        operand=None
    )
    # calculate the magnetic field at the particle positions

    boris_vmap = jax.vmap(boris_single_particle, in_axes=(0, 0, 0, 0, 0, 0, 0, 0, 0, None, None, None, None))
    newvx, newvy, newvz = boris_vmap(vx, vy, vz, efield_atx, efield_aty, efield_atz, bfield_atx, bfield_aty, bfield_atz, q, m, dt, constants)
    # apply the Boris algorithm to update the velocities of the particles

    particles.set_velocity(newvx, newvy, newvz)
    # set the new velocities of the particles
    return particles

@jit
def boris_single_particle(vx, vy, vz, efield_atx, efield_aty, efield_atz, bfield_atx, bfield_aty, bfield_atz, q, m, dt, constants):
    """
    Updates the velocity of a single particle using the Boris algorithm.
    Args:
        x (float): Initial x position of the particle.
        y (float): Initial y position of the particle.
        z (float): Initial z position of the particle.
        vx (float): Initial x component of the particle's velocity.
        vy (float): Initial y component of the particle's velocity.
        vz (float): Initial z component of the particle's velocity.
        efield_atx (float): x component of the electric field at the particle's position.
        efield_aty (float): y component of the electric field at the particle's position.
        efield_atz (float): z component of the electric field at the particle's position.
        bfield_atx (float): x component of the magnetic field at the particle's position.
        bfield_aty (float): y component of the magnetic field at the particle's position.
        bfield_atz (float): z component of the magnetic field at the particle's position.
        q (float): Charge of the particle.
        m (float): Mass of the particle.
        dt (float): Time step for the update.
    Returns:
        tuple: Updated velocity components (vx, vy, vz) of the particle.
    """

    v = jnp.array([vx, vy, vz])
    # convert v into an array

    vminus = v + q*dt/(2*m)*jnp.array([efield_atx, efield_aty, efield_atz])
    # get v minus vector

    t = q*dt/(2*m)*jnp.array([bfield_atx, bfield_aty, bfield_atz])
    # calculate the t vector
    vprime = vminus + jnp.cross(vminus, t)
    # calculate the v prime vector

    s = 2*t / (1 + t[0]**2 + t[1]**2 + t[2]**2)
    # calculate the s vector
    vplus = vminus + jnp.cross(vprime, s)
    # calculate the v plus vector

    newv = vplus + q*dt/(2*m)*jnp.array([efield_atx, efield_aty, efield_atz])
    # calculate the new velocity
    return newv[0], newv[1], newv[2]



@jit
def relativistic_boris_single_particle(vx, vy, vz, efield_atx, efield_aty, efield_atz, bfield_atx, bfield_aty, bfield_atz, q, m, dt, constants):
    """
    Perform a single step of the relativistic Boris algorithm for a charged particle.

    This function calculates the updated velocity of a charged particle under the influence
    of electric and magnetic fields using the relativistic Boris algorithm. The algorithm
    ensures energy conservation and is widely used in particle-in-cell (PIC) simulations.

    Args:
        vx (float): Initial velocity of the particle in the x-direction (m/s).
        vy (float): Initial velocity of the particle in the y-direction (m/s).
        vz (float): Initial velocity of the particle in the z-direction (m/s).
        efield_atx (float): Electric field at the particle's position in the x-direction (V/m).
        efield_aty (float): Electric field at the particle's position in the y-direction (V/m).
        efield_atz (float): Electric field at the particle's position in the z-direction (V/m).
        bfield_atx (float): Magnetic field at the particle's position in the x-direction (T).
        bfield_aty (float): Magnetic field at the particle's position in the y-direction (T).
        bfield_atz (float): Magnetic field at the particle's position in the z-direction (T).
        q (float): Charge of the particle (Coulombs).
        m (float): Mass of the particle (kg).
        dt (float): Time step for the simulation (seconds).
        constants (dict): Dictionary containing physical constants. Must include:
            - 'C': Speed of light in vacuum (m/s).

    Returns:
        tuple: Updated velocity components of the particle in the x, y, and z directions (m/s).
    """


    C = constants['C']
    # speed of light

    v = jnp.array([vx, vy, vz])
    # convert v into an array

    gamma = jnp.sqrt( 1  + (  (v[0]**2 + v[1]**2 + v[2]**2) / C**2 ) )
    # define the gamma factor

    vminus = v * gamma + q*dt/(2*m)*jnp.array([efield_atx, efield_aty, efield_atz])
    # get v minus vector

    gamma_minus = jnp.sqrt( 1  + (  (vminus[0]**2 + vminus[1]**2 + vminus[2]**2) / C**2 ) )

    t = q*dt/(2*m)*jnp.array([bfield_atx, bfield_aty, bfield_atz]) / gamma_minus
    # calculate the t vector
    vprime = vminus + jnp.cross(vminus, t)
    # calculate the v prime vector

    s = 2*t / (1 + t[0]**2 + t[1]**2 + t[2]**2)
    # calculate the s vector
    vplus = vminus + jnp.cross(vprime, s)
    # calculate the v plus vector

    newv = vplus + q*dt/(2*m)*jnp.array([efield_atx, efield_aty, efield_atz])
    # calculate the new velocity

    new_gamma = jnp.sqrt( 1  +  (  (newv[0]**2 + newv[1]**2 + newv[2]**2) / C**2 ) )
    # define the new gamma factor

    return newv[0] / new_gamma, newv[1] / new_gamma, newv[2] / new_gamma



def create_trilinear_interpolator(field, grid, periodic=True):
    """
    Create a trilinear interpolation function for a given 3D field and grid.

    Handles cases where any dimension (nx, ny, nz) has only 1 grid point.
    In such cases, no interpolation is performed along that dimension.

    Args:
        field (ndarray): The 3D field to interpolate.
        grid (tuple): A tuple of three arrays representing the grid points in the x, y, and z directions.
        periodic (tuple): A tuple of three booleans indicating whether each dimension is periodic.
                         Default is (True, True, True) for fully periodic domains.

    Returns:
        function: A function that takes (x, y, z) coordinates and returns the interpolated values.
    """
    x_grid, y_grid, z_grid = grid

    dx = x_grid[1] - x_grid[0]
    dy = y_grid[1] - y_grid[0]
    dz = z_grid[1] - z_grid[0]
    # calculate the grid spacing in each direction

    x_min, x_max = x_grid[0], x_grid[-1]
    y_min, y_max = y_grid[0], y_grid[-1]
    z_min, z_max = z_grid[0], z_grid[-1]
    # get grid bounds

    x_wind = x_max - x_min
    y_wind = y_max - y_min
    z_wind = z_max - z_min
    # get spatial widths of the grid

    Nx = len(x_grid)
    Ny = len(y_grid)
    Nz = len(z_grid)
    # get the number of grid points in each direction

    @jit
    def interpolator(x, y, z):

        # Handle periodic boundaries by wrapping coordinates
        if periodic:
            x = x_min + jnp.mod(x - x_min, x_wind)
            y = y_min + jnp.mod(y - y_min, y_wind)
            z = z_min + jnp.mod(z - z_min, z_wind)

        # Convert coordinates to grid indices (fractional)
        # Handle single-point dimensions specially
        xi = jnp.where(Nx == 1, 0.0, (x - x_min) / dx)
        yi = jnp.where(Ny == 1, 0.0, (y - y_min) / dy)
        zi = jnp.where(Nz == 1, 0.0, (z - z_min) / dz)

        # Find the lower-left-bottom corner indices
        x0 = jnp.where(Nx == 1, 0, jnp.floor(xi).astype(int))
        y0 = jnp.where(Ny == 1, 0, jnp.floor(yi).astype(int))
        z0 = jnp.where(Nz == 1, 0, jnp.floor(zi).astype(int))

        # Handle boundary conditions
        if periodic:
            x0 = jnp.where(Nx == 1, 0, jnp.mod(x0, Nx))
            y0 = jnp.where(Ny == 1, 0, jnp.mod(y0, Ny))
            z0 = jnp.where(Nz == 1, 0, jnp.mod(z0, Nz))
            x1 = jnp.where(Nx == 1, 0, jnp.mod(x0 + 1, Nx))
            y1 = jnp.where(Ny == 1, 0, jnp.mod(y0 + 1, Ny))
            z1 = jnp.where(Nz == 1, 0, jnp.mod(z0 + 1, Nz))
        else:
            x0 = jnp.where(Nx == 1, 0, jnp.clip(x0, 0, Nx - 2))
            y0 = jnp.where(Ny == 1, 0, jnp.clip(y0, 0, Ny - 2))
            z0 = jnp.where(Nz == 1, 0, jnp.clip(z0, 0, Nz - 2))
            x1 = jnp.where(Nx == 1, 0, jnp.clip(x0 + 1, 0, Nx - 1))
            y1 = jnp.where(Ny == 1, 0, jnp.clip(y0 + 1, 0, Ny - 1))
            z1 = jnp.where(Nz == 1, 0, jnp.clip(z0 + 1, 0, Nz - 1))

        # Calculate the fractional parts (weights)
        # For single-point dimensions, weight is meaningless so set to 0
        wx = jnp.where(Nx == 1, 0.0, xi - jnp.floor(xi))
        wy = jnp.where(Ny == 1, 0.0, yi - jnp.floor(yi))
        wz = jnp.where(Nz == 1, 0.0, zi - jnp.floor(zi))

        # Trilinear interpolation
        c000 = field[x0, y0, z0]
        c001 = field[x0, y0, z1]
        c010 = field[x0, y1, z0]
        c011 = field[x0, y1, z1]
        c100 = field[x1, y0, z0]
        c101 = field[x1, y0, z1]
        c110 = field[x1, y1, z0]
        c111 = field[x1, y1, z1]

        # Interpolate along x-axis first
        c00 = c000 * (1 - wx) + c100 * wx
        c01 = c001 * (1 - wx) + c101 * wx
        c10 = c010 * (1 - wx) + c110 * wx
        c11 = c011 * (1 - wx) + c111 * wx

        # Then interpolate along y-axis
        c0 = c00 * (1 - wy) + c10 * wy
        c1 = c01 * (1 - wy) + c11 * wy

        # Finally interpolate along z-axis
        value = c0 * (1 - wz) + c1 * wz

        return value

    vmap_interpolator = jax.vmap(interpolator, in_axes=(0, 0, 0), out_axes=0)
    # vectorize the interpolator for batch processing

    return vmap_interpolator

def create_quadratic_interpolator(field, grid, periodic=True):
    """
    Create a quadratic interpolation function for a given 3D field and grid.

    Args:
        field (ndarray): The 3D field to interpolate.
        grid (tuple): A tuple of three arrays representing the grid points in the x, y, and z directions.
        periodic (tuple): A tuple of three booleans indicating whether each dimension is periodic.
                         Default is (True, True, True) for fully periodic domains.

    Returns:
        function: A function that takes (x, y, z) coordinates and returns the interpolated values.
    """
    x_grid, y_grid, z_grid = grid

    # Calculate grid spacing and bounds
    dx = x_grid[1] - x_grid[0] if len(x_grid) > 1 else 1.0
    dy = y_grid[1] - y_grid[0] if len(y_grid) > 1 else 1.0
    dz = z_grid[1] - z_grid[0] if len(z_grid) > 1 else 1.0

    x_min, x_max = x_grid[0], x_grid[-1]
    y_min, y_max = y_grid[0], y_grid[-1]
    z_min, z_max = z_grid[0], z_grid[-1]

    # Calculate domain widths
    x_width = x_max - x_min
    y_width = y_max - y_min
    z_width = z_max - z_min

    Nx = len(x_grid)
    Ny = len(y_grid)
    Nz = len(z_grid)

    @jit
    def interpolator(x, y, z):
        # Handle periodic boundaries by wrapping coordinates
        if periodic:
            x = x_min + jnp.mod(x - x_min, x_width)
            y = y_min + jnp.mod(y - y_min, y_width)
            z = z_min + jnp.mod(z - z_min, z_width)

        # Handle each dimension separately to account for different grid sizes
        def get_indices_and_points(coord, grid_1d, coord_min, d_spacing):
            grid_len = len(grid_1d)

            # Convert coordinate to fractional grid index
            if grid_len == 1:
                frac_idx = 0.0
            else:
                frac_idx = (coord - coord_min) / d_spacing

            # Use JAX-compatible conditionals
            idx_base = jnp.floor(frac_idx).astype(int) if grid_len > 1 else 0

            # For grids with 3+ points, use quadratic interpolation
            idx_quad = jnp.clip(idx_base, 1, grid_len - 2)

            # For grids with 2 points, use linear interpolation
            idx_lin = jnp.clip(idx_base, 0, 0)

            # For single-point grids, use index 0
            idx_single = 0

            # Select appropriate index based on grid size
            idx = jnp.where(grid_len >= 3, idx_quad,
                   jnp.where(grid_len == 2, idx_lin, idx_single))

            # Handle periodic boundary conditions for indices
            if periodic and grid_len > 1:
                # For periodic boundaries, wrap the indices
                idx_m1 = jnp.mod(idx - 1, grid_len)
                idx_0 = jnp.mod(idx, grid_len)
                idx_p1 = jnp.mod(idx + 1, grid_len)

                p0 = jnp.where(grid_len >= 3, grid_1d[idx_m1],
                      jnp.where(grid_len == 2, grid_1d[0], grid_1d[0]))
                p1 = jnp.where(grid_len >= 2, grid_1d[idx_0], grid_1d[0])
                p2 = jnp.where(grid_len >= 3, grid_1d[idx_p1],
                      jnp.where(grid_len == 2, grid_1d[1], grid_1d[0]))
            else:
                # For non-periodic boundaries, use clipping
                # Get the three points for interpolation
                # For 3+ points: normal stencil
                # For 2 points: duplicate last point
                # For 1 point: duplicate the single point
                p0 = jnp.where(grid_len >= 3, grid_1d[jnp.clip(idx - 1, 0, grid_len - 1)],
                      jnp.where(grid_len == 2, grid_1d[0], grid_1d[0]))

                p1 = jnp.where(grid_len >= 2, grid_1d[jnp.clip(idx, 0, grid_len - 1)], grid_1d[0])

                p2 = jnp.where(grid_len >= 3, grid_1d[jnp.clip(idx + 1, 0, grid_len - 1)],
                      jnp.where(grid_len == 2, grid_1d[1], grid_1d[0]))

            return idx, p0, p1, p2

        x_idx, x0, x1, x2 = get_indices_and_points(x, x_grid, x_min, dx)
        y_idx, y0, y1, y2 = get_indices_and_points(y, y_grid, y_min, dy)
        z_idx, z0, z1, z2 = get_indices_and_points(z, z_grid, z_min, dz)

        def quadratic_weights(t, t0, t1, t2):
            # Handle degenerate cases (when points are equal)
            eps = 1e-12

            # Check for degenerate cases using JAX-compatible logic
            all_same = (jnp.abs(t0 - t1) < eps) & (jnp.abs(t1 - t2) < eps)
            t0_eq_t1 = (jnp.abs(t0 - t1) < eps) & (jnp.abs(t1 - t2) >= eps)
            t1_eq_t2 = (jnp.abs(t0 - t1) >= eps) & (jnp.abs(t1 - t2) < eps)

            # Standard quadratic weights
            denom0 = (t0 - t1) * (t0 - t2)
            denom1 = (t1 - t0) * (t1 - t2)
            denom2 = (t2 - t0) * (t2 - t1)

            # Avoid division by zero by adding small epsilon where needed
            denom0 = jnp.where(jnp.abs(denom0) < eps, eps, denom0)
            denom1 = jnp.where(jnp.abs(denom1) < eps, eps, denom1)
            denom2 = jnp.where(jnp.abs(denom2) < eps, eps, denom2)

            w0_standard = (t - t1) * (t - t2) / denom0
            w1_standard = (t - t0) * (t - t2) / denom1
            w2_standard = (t - t0) * (t - t1) / denom2

            # Linear interpolation weights for degenerate cases
            w_lin_01 = jnp.where(jnp.abs(t1 - t0) < eps, 0.0, (t - t0) / (t1 - t0))
            w_lin_12 = jnp.where(jnp.abs(t2 - t1) < eps, 0.0, (t - t1) / (t2 - t1))

            # Select weights based on degenerate cases
            w0 = jnp.where(all_same, 0.0,
                  jnp.where(t0_eq_t1, 0.0,
                   jnp.where(t1_eq_t2, 1.0 - w_lin_01, w0_standard)))

            w1 = jnp.where(all_same, 1.0,
                  jnp.where(t0_eq_t1, 1.0 - w_lin_12,
                   jnp.where(t1_eq_t2, w_lin_01, w1_standard)))

            w2 = jnp.where(all_same, 0.0,
                  jnp.where(t0_eq_t1, w_lin_12,
                   jnp.where(t1_eq_t2, 0.0, w2_standard)))

            return w0, w1, w2

        wx0, wx1, wx2 = quadratic_weights(x, x0, x1, x2)
        wy0, wy1, wy2 = quadratic_weights(y, y0, y1, y2)
        wz0, wz1, wz2 = quadratic_weights(z, z0, z1, z2)

        interpolated_value = 0.0
        for i, wx in enumerate([wx0, wx1, wx2]):
            for j, wy in enumerate([wy0, wy1, wy2]):
                for k, wz in enumerate([wz0, wz1, wz2]):
                    # Calculate array indices with proper bounds checking
                    # Use JAX-compatible conditionals

                    # X index calculation
                    if periodic and Nx > 1:
                        xi_quad = jnp.mod(x_idx - 1 + i, Nx)  # For 3+ points with periodic
                        xi_lin = jnp.mod(x_idx + i - 1, Nx)   # For 2 points with periodic
                    else:
                        xi_quad = x_idx - 1 + i  # For 3+ points
                        xi_lin = jnp.clip(x_idx + i - 1, 0, max(1, Nx - 1))  # For 2 points
                    xi_single = 0  # For 1 point

                    xi = jnp.where(Nx >= 3, xi_quad,
                          jnp.where(Nx == 2, xi_lin, xi_single))

                    # Y index calculation
                    if periodic and Ny > 1:
                        yi_quad = jnp.mod(y_idx - 1 + j, Ny)
                        yi_lin = jnp.mod(y_idx + j - 1, Ny)
                    else:
                        yi_quad = y_idx - 1 + j
                        yi_lin = jnp.clip(y_idx + j - 1, 0, max(1, Ny - 1))
                    yi_single = 0

                    yi = jnp.where(Ny >= 3, yi_quad,
                          jnp.where(Ny == 2, yi_lin, yi_single))

                    # Z index calculation
                    if periodic and Nz > 1:
                        zi_quad = jnp.mod(z_idx - 1 + k, Nz)
                        zi_lin = jnp.mod(z_idx + k - 1, Nz)
                    else:
                        zi_quad = z_idx - 1 + k
                        zi_lin = jnp.clip(z_idx + k - 1, 0, max(1, Nz - 1))
                    zi_single = 0

                    zi = jnp.where(Nz >= 3, zi_quad,
                          jnp.where(Nz == 2, zi_lin, zi_single))

                    # Final bounds check (only needed for non-periodic)
                    if not periodic:
                        xi = jnp.clip(xi, 0, field.shape[0] - 1)
                        yi = jnp.clip(yi, 0, field.shape[1] - 1)
                        zi = jnp.clip(zi, 0, field.shape[2] - 1)

                    interpolated_value += wx * wy * wz * field[xi, yi, zi]

        return interpolated_value

    vmap_interpolator = jax.vmap(interpolator, in_axes=(0, 0, 0), out_axes=0)
    return vmap_interpolator

@jit
def wrap_around(ix, size):
    """Wrap around index to ensure it is within bounds."""
    return jax.lax.cond(
        ix > size - 1,
        lambda _: ix - size,
        lambda _: ix,
        operand=None
    )


import jax
from jax.experimental import mesh_utils
from jax.sharding import Mesh, PartitionSpec, NamedSharding
import os
import functools
from functools import partial
import toml
import matplotlib.pyplot as plt
import jax.numpy as jnp
#from memory_profiler import profile

from PyPIC3D.particle import (
    load_particles_from_toml
)

from PyPIC3D.utils import (
    courant_condition,
    update_parameters_from_toml,
    build_yee_grid, convert_to_jax_compatible, load_external_fields_from_toml,
    print_stats, particle_sanity_check, build_plasma_parameters_dict,
    make_dir, compute_energy, build_collocated_grid
)

from PyPIC3D.fields import (
    calculateE, initialize_fields
)

from PyPIC3D.pstd import (
    spectral_curl
)

from PyPIC3D.fdtd import (
    centered_finite_difference_curl
)


from PyPIC3D.plotting import (
    plot_initial_histograms
)


from PyPIC3D.evolve import (
    time_loop_electrodynamic, time_loop_electrostatic, time_loop_vector_potential
)

from PyPIC3D.J import (
    J_from_rhov, Esirkepov_current, VB_current
)
from PyPIC3D.vector_potential import initialize_vector_potential


def default_parameters():
    """
    Returns a dictionary of default parameters for the simulation.

    Returns:
    dict: A dictionary of default parameters.
    """
    plotting_parameters = {
    "plotting" : True,
    "save_data": False,
    "plotfields": False,
    "plotpositions": False,
    "plotvelocities": False,
    "plotenergy": True,
    "plotcurrent": False,
    "plasmaFreq": False,
    "phaseSpace": False,
    "plot_errors": False,
    "plot_dispersion": False,
    'plot_chargeconservation': False,
    "plot_vtk_particles": True,
    "plotting_interval": 10
    }
    # dictionary for plotting/saving data

    simulation_parameters = {
        "name": "Default Simulation",
        "output_dir": ".",
        "solver": "spectral",  # solver: spectral, fdtd, vector_potential
        "bc": "spectral",  # boundary conditions: periodic, dirichlet, neumann
        "Nx": 30,  # number of array spacings in x
        "Ny": 30,  # number of array spacings in y
        "Nz": 30,  # number of array spacings in z
        "x_wind": 1e-2,  # size of the spatial window in x in meters
        "y_wind": 1e-2,  # size of the spatial window in y in meters
        "z_wind": 1e-2,  # size of the spatial window in z in meters
        "t_wind": 1e-12,  # size of the temporal window in seconds
        "dt": None,  # time step in seconds
        "Nt": None,  # number of time steps
        "electrostatic": False,  # boolean for electrostatic simulation
        "benchmark": False, # boolean for using the profiler
        "verbose": False, # boolean for printing verbose output
        "GPUs": False, # boolean for using GPUs
        "ncores": 4, # number of cores to use
        "ncpus": 1, # number of CPUs to use
        "cfl"  : 1, # CFL condition number
        "ds_per_debye" : None, # number of grid spacings per debye length
        "shape_factor" : 1, # shape factor for the simulation (1 for 1st order, 2 for 2nd order)
        "current_calculation": "j_from_rhov",  # current calculation method: esirkepov, villasenor_buneman, j_from_rhov
    }
    # dictionary for simulation parameters

    constants = {
        "eps": 8.85418782e-12,  # permitivity of freespace
        "mu" : 1.25663706e-6, # permeability of free space
        "C": 2.99792458e8,  # Speed of light in m/s
        "kb": 1.380649e-23,  # Boltzmann's constant in J/K
    }

    return plotting_parameters, simulation_parameters, constants
    # return the dictionaries


def setup_write_dir(simulation_parameters, plotting_parameters):
        output_dir = simulation_parameters['output_dir']
        # get the output directory from the simulation parameters
        make_dir(f'{output_dir}/data')
        # make the directory for the data


#@profile
def initialize_simulation(toml_file):
    """
    Initializes the simulation environment based on the provided TOML configuration file.

    Args:
        toml_file (str): Path to the TOML configuration file. If None, default parameters are used.

    Returns:
        tuple: A tuple containing the following elements:
            evolve_loop (function): The function to evolve the simulation loop.
            particles (list): List of particle objects initialized from the configuration file.
            Ex, Ey, Ez (jax.numpy.ndarray): Electric field components.
            Bx, By, Bz (jax.numpy.ndarray): Magnetic field components.
            Jx, Jy, Jz (jax.numpy.ndarray): Current density components.
            phi (jax.numpy.ndarray): Electric potential.
            rho (jax.numpy.ndarray): Charge density.
            E_grid, B_grid (dict): Grids for electric and magnetic fields.
            world (dict): Dictionary containing world parameters such as spatial resolution and domain size.
            simulation_parameters (dict): Dictionary containing simulation parameters.
            constants (dict): Dictionary containing physical constants.
            plotting_parameters (dict): Dictionary containing parameters for plotting.
            plasma_parameters (dict): Dictionary containing plasma parameters.
            M (jax.numpy.ndarray or None): Preconditioner matrix, if neural network preconditioning is used.
            solver (str): Solver type, either "spectral" or "fdtd".
            bc (str): Boundary conditions.
            electrostatic (bool): Flag indicating if the simulation is electrostatic.
            verbose (bool): Flag indicating if verbose output is enabled.
            GPUs (int): Number of GPUs to use.
            start (float): Start time of the simulation initialization.
            Nt (int): Number of time steps.
            curl_func (function): Function to compute the curl of the fields.
            pecs (list): List of perfectly electrical conductor boundaries.
            lasers (list): List of laser objects initialized from the configuration file.
            surfaces (list): List of material surfaces initialized from the configuration file.
    """

    plotting_parameters, simulation_parameters, constants = default_parameters()
    # load the default parameters

    if toml_file is not None:
        simulation_parameters, plotting_parameters, constants = update_parameters_from_toml(toml_file, simulation_parameters, plotting_parameters, constants)

    print(f"Initializing Simulation: { simulation_parameters['name'] }\n")

    x_wind, y_wind, z_wind = simulation_parameters['x_wind'], simulation_parameters['y_wind'], simulation_parameters['z_wind']
    Nx, Ny, Nz = simulation_parameters['Nx'], simulation_parameters['Ny'], simulation_parameters['Nz']
    t_wind = simulation_parameters['t_wind']
    electrostatic = simulation_parameters['electrostatic']
    solver = simulation_parameters['solver']
    bc = simulation_parameters['bc']
    verbose = simulation_parameters['verbose']
    GPUs = simulation_parameters['GPUs']
    #ncores = simulation_parameters['ncores']
    ncpus = simulation_parameters['ncpus']
    # set the simulation parameters

    if 'ncores' in simulation_parameters:
        os.environ["XLA_FLAGS"] = f'--xla_force_host_platform_device_count={simulation_parameters['ncores']}'
    # set the number of cores to use

    setup_write_dir(simulation_parameters, plotting_parameters)
    # setup the write directory

    dx, dy, dz = x_wind/Nx, y_wind/Ny, z_wind/Nz
    # compute the spatial resolution

    if simulation_parameters['dt'] is not None:
        print(f"Using user defined dt: {simulation_parameters['dt']}")
        dt = simulation_parameters['dt']
    else:
        courant_number = simulation_parameters['cfl']
        dt = courant_condition(courant_number, dx, dy, dz, simulation_parameters, constants)
    # compute the time step
    if simulation_parameters['Nt'] is not None:
        Nt = simulation_parameters['Nt']
    else:
        Nt     = int( t_wind / dt )
    # Nt for resolution
    world = {'dt': dt, 'Nt': Nt, 'dx': dx, 'dy': dy, 'dz': dz, 'Nx': Nx, 'Ny': Ny, 'Nz': Nz, 'x_wind': x_wind, 'y_wind': y_wind, 'z_wind': z_wind}
    # set the simulation world parameters

    world = convert_to_jax_compatible(world)
    constants = convert_to_jax_compatible(constants)
    simulation_parameters = convert_to_jax_compatible(simulation_parameters)
    plotting_parameters = convert_to_jax_compatible(plotting_parameters)
    # convert the world parameters to jax compatible format

    if solver == "vector_potential":
        B_grid, E_grid = build_collocated_grid(world)
        # build the grid for the fields
    else:
        B_grid, E_grid = build_yee_grid(world)
        # build the Yee grid for the fields

    if not os.path.exists(f"{simulation_parameters['output_dir']}/data"):
        os.makedirs(f"{simulation_parameters['output_dir']}/data")
        # create the data directory if it doesn't exist
    ################################### INITIALIZE PARTICLES AND FIELDS ########################################################

    particles = load_particles_from_toml(toml_file, simulation_parameters, world, constants)
    # load the particles from the configuration file

    for species in particles:
        name = species.get_name()
        name = name.replace(" ", "_")
        # replace spaces with underscores in the name
        plot_initial_histograms(species, world, path=f"{simulation_parameters['output_dir']}/data", name=name)
        # plot the initial histograms of the particles

    print_stats(world)
    # print the statistics of the simulation

    if particles:
        plasma_parameters = build_plasma_parameters_dict(world, constants, particles[0], dt)
        # build the plasma parameters dictionary
    else:
        plasma_parameters = {}
        # if no particles are loaded, set plasma parameters to empty dictionary

    particle_sanity_check(particles)
    # ensure the arrays for the particles are of the correct shape

    E, B, J, phi, rho = initialize_fields(Nx, Ny, Nz)
    # initialize the electric and magnetic fields

    # load any external fields
    fields = [component for field in [E, B, J] for component in field]
    # convert the E, B, and J tuples into one big list
    fields = load_external_fields_from_toml(fields, toml_file)
    # add any external fields to the simulation
    E, B, J = fields[:3], fields[3:6], fields[6:9]
    # convert the fields list back into tuples

    if solver == "spectral":
        curl_func = functools.partial(spectral_curl, world=world)
    else:
        curl_func = functools.partial(centered_finite_difference_curl, dx=dx, dy=dy, dz=dz, bc=bc)


    ######################### COMPUTE INITIAL ENERGY ########################################################
    e_energy, b_energy, kinetic_energy = compute_energy(particles, E, B, world, constants)
    # compute the initial energy of the system
    print(f"Initial Electric Field Energy: {e_energy:.2e} J")
    print(f"Initial Magnetic Field Energy: {b_energy:.2e} J")
    print(f"Initial Kinetic Energy: {kinetic_energy:.2e} J")
    print(f"Total Initial Energy: {e_energy + b_energy + kinetic_energy:.2e} J\n")
    # print the initial energy of the system

    if electrostatic:
        print("Using electrostatic solver\n")
        evolve_loop = time_loop_electrostatic

    elif solver == "vector_potential":
        print("Using vector potential solver\n")
        evolve_loop = time_loop_vector_potential

    else:
        print(f"Using electrodynamic solver with: {solver}\n")
        evolve_loop = time_loop_electrodynamic
    # set the evolve loop function based on the electrostatic flag

    if simulation_parameters['current_calculation'] == "esirkepov":
        print("Using Esirkepov current calculation method\n")
        J_func = Esirkepov_current
    elif simulation_parameters['current_calculation'] == "villasenor_buneman":
        print("Using Villasenor-Buneman current calculation method\n")
        J_func = VB_current
    elif simulation_parameters['current_calculation'] == "j_from_rhov":
        print("Using J from rhov current calculation method\n")
        J_func = J_from_rhov


    if solver == "vector_potential":
        A2, A1, A0 = initialize_vector_potential(J, world, constants)
        # initialize the vector potential A based on the current density J
        fields = (E, B, J, rho, phi, A2, A1, A0)
    else:
        fields = (E, B, J, rho, phi)

    return evolve_loop, particles, fields, E_grid, B_grid, world, simulation_parameters, constants, plotting_parameters, plasma_parameters, \
        solver, bc, electrostatic, verbose, GPUs, Nt, curl_func, J_func



import jax
import jax.numpy as jnp

from PyPIC3D.fdtd import (
    centered_finite_difference_curl,
    centered_finite_difference_laplacian,
    centered_finite_difference_divergence,
    centered_finite_difference_gradient
)


def initialize_vector_potential(J, world, constants):
    """
    Initialize the vector potential A based on the current density J.

    Args:
        J (tuple): Current density components (Jx, Jy, Jz).
        world (dict): Simulation world parameters including 'dx', 'dy', 'dz'.
        constants (dict): Physical constants including 'mu'.

    Returns:
        tuple: Initial vector potential components (Ax, Ay, Az).
    """
    Jx, Jy, Jz = J

    Ax = jnp.zeros_like(Jx)
    Ay = jnp.zeros_like(Jy)
    Az = jnp.zeros_like(Jz)
    # initialize A as zero

    A = Ax, Ay, Az

    return A, A, A

@jax.jit
def update_vector_potential(J, world, constants, A1, A0):
    """
    Update the vector potential A based on the magnetic field B using centered finite difference.

    Args:
        A (tuple): Vector potential components (Ax, Ay, Az).
        J (tuple): Current density components (Jx, Jy, Jz).
        dt (float): Time step.
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.
        bc (str): Boundary condition type ('periodic', 'neumann', 'dirichlet').

    Returns:
        tuple: Updated vector potential components (Ax_new, Ay_new, Az_new).
    """
    Ax0, Ay0, Az0 = A0
    Ax, Ay, Az = A1
    Jx, Jy, Jz = J

    dx, dy, dz = world['dx'], world['dy'], world['dz']

    mu = constants['mu']
    C  = constants['C']
    dt = world['dt']

    laplacian_Ax = centered_finite_difference_laplacian(Ax, dx, dy, dz, 'periodic')
    laplacian_Ay = centered_finite_difference_laplacian(Ay, dx, dy, dz, 'periodic')
    laplacian_Az = centered_finite_difference_laplacian(Az, dx, dy, dz, 'periodic')
    # calculate the Laplacian of the vector potential using centered finite difference

    divA         = centered_finite_difference_divergence(Ax, Ay, Az, dx, dy, dz, 'periodic')
    # calculate the divergence of the vector potential using centered finite difference

    gradx, grady, gradz = centered_finite_difference_gradient(divA, dx, dy, dz, 'periodic')
    # calculate the gradient of the divergence of the vector potential using centered finite difference

    Ax_new = 2 * Ax - Ax0 + C**2 * dt**2 * ( mu * Jx  + laplacian_Ax + gradx )
    Ay_new = 2 * Ay - Ay0 + C**2 * dt**2 * ( mu * Jy  + laplacian_Ay + grady )
    Az_new = 2 * Az - Az0 + C**2 * dt**2 * ( mu * Jz  + laplacian_Az + gradz )
    # update the vector potential using centered finite difference

    return Ax_new, Ay_new, Az_new

@jax.jit
def E_from_A(A2, A0, world):
    Ax, Ay, Az = A2
    Ax0, Ay0, Az0 = A0
    dt = world['dt']

    Ex = -1 * (Ax - Ax0) / (2 * dt )
    Ey = -1 * (Ay - Ay0) / (2 * dt )
    Ez = -1 * (Az - Az0) / (2 * dt )
    # calculate the electric field from the vector potential using centered finite difference

    return Ex, Ey, Ez

@jax.jit
def B_from_A(A, world):
    """
    Calculate the magnetic field B from the vector potential A using centered finite difference.

    Args:
        A (tuple): Vector potential components (Ax, Ay, Az).
        world (dict): Simulation world parameters including 'dx', 'dy', 'dz'.

    Returns:
        tuple: Magnetic field components (Bx, By, Bz).
    """
    Ax, Ay, Az = A
    dx, dy, dz = world['dx'], world['dy'], world['dz']

    Bx, By, Bz = centered_finite_difference_curl(Ax, Ay, Az, dx, dy, dz, 'periodic')

    return Bx, By, Bz


# Christopher Woolford July 2024
# I am writing a 3D PIC code in Python using the Jax library
# to test the feasibility of using Jax for plasma simulations to take advantage of
# Jax's auto-differentiation capabilities

########################################## IMPORT LIBRARIES #############################################
from sys import path
import time
import jax
from jax import block_until_ready
import jax.numpy as jnp
from tqdm import tqdm
#from memory_profiler import profile
# Importing relevant libraries

from PyPIC3D.plotting import (
    plotter, write_particles_phase_space, write_data, plot_vtk_particles, plot_field_slice_vtk
)

from PyPIC3D.utils import (
    dump_parameters_to_toml, load_config_file, compute_energy
)

from PyPIC3D.initialization import (
    initialize_simulation
)


# Importing functions from the PyPIC3D package
############################################################################################################


def run_PyPIC3D(config_file):
    ##################################### INITIALIZE SIMULATION ################################################

    loop, particles, fields, E_grid, B_grid, world, simulation_parameters, constants, plotting_parameters, \
            plasma_parameters, solver, bc, electrostatic, verbose, GPUs, Nt, curl_func, J_func = initialize_simulation(config_file)
    # initialize the simulation

    jit_loop = jax.jit(loop, static_argnames=('curl_func', 'J_func','solver', 'bc'))

    dt = world['dt']
    output_dir = simulation_parameters['output_dir']

    field_names = ["E_magnitude", "B_magnitude"]

    ############################################################################################################

    ###################################################### SIMULATION LOOP #####################################

    for t in tqdm(range(Nt)):
        # plotter(t, particles, E, B, J, rho, phi, E_grid, B_grid, world, constants, plotting_parameters, simulation_parameters)
        # plot the data
        if t % plotting_parameters['plotting_interval'] == 0:
            E, B, J, *rest = fields
            # unpack the fields
            e_energy, b_energy, kinetic_energy = compute_energy(particles, E, B, world, constants)
            # Compute the energy of the system
            write_data(f"{output_dir}/data/total_energy.txt", t * dt, e_energy + b_energy + kinetic_energy)
            write_data(f"{output_dir}/data/electric_field_energy.txt", t * dt, e_energy)
            write_data(f"{output_dir}/data/magnetic_field_energy.txt", t * dt, b_energy)
            write_data(f"{output_dir}/data/kinetic_energy.txt", t * dt, kinetic_energy)
            # Write the total energy to a file
            total_momentum = sum(particle_species.momentum() for particle_species in particles)
            # Total momentum of the particles
            write_data(f"{output_dir}/data/total_momentum.txt", t * dt, total_momentum)
            # Write the total momentum to a file

            write_particles_phase_space(particles, t, output_dir)

            E_magnitude = jnp.sqrt(E[0]**2 + E[1]**2 + E[2]**2)[:,:,world['Nz']//2]
            B_magnitude = jnp.sqrt(B[0]**2 + B[1]**2 + B[2]**2)[:,:,world['Nz']//2]
            fields_mag = [E_magnitude, B_magnitude]
            plot_field_slice_vtk(fields_mag, field_names, 2, E_grid, t, "fields", output_dir, world)
            # Plot the fields in VTK format

            # if plotting_parameters['plot_vtk_particles']:
                # plot_vtk_particles(particles, t, output_dir)
            # Plot the particles in VTK format

        particles, fields = jit_loop(particles, fields, E_grid, B_grid, world, constants, curl_func, J_func, solver, bc)
        # time loop to update the particles and fields

    return Nt, plotting_parameters, simulation_parameters, plasma_parameters, constants, particles, fields, world

def main():
    ###################### JAX SETTINGS ########################################################################
    jax.config.update("jax_enable_x64", True)
    # set Jax to use 64 bit precision
    jax.config.update("jax_debug_nans", True)
    # debugging for nans
    jax.config.update('jax_platform_name', 'cpu')
    # set Jax to use CPUs
    #jax.config.update("jax_disable_jit", True)
    ############################################################################################################

    toml_file = load_config_file()
    # load the configuration file

    start = time.time()
    # start the timer

    Nt, plotting_parameters, simulation_parameters, plasma_parameters, constants, particles, fields, world =  block_until_ready(run_PyPIC3D(toml_file))
    # run the PyPIC3D simulation

    end = time.time()
    # end the timer

    E, B, J, *rest = fields
    # unpack the fields

    e_energy, b_energy, kinetic_energy = compute_energy(particles, E, B, world, constants)
    # compute the energy of the system
    print(f"Final Electric Field Energy: {e_energy}")
    print(f"Final Magnetic Field Energy: {b_energy}")
    print(f"Final Kinetic Energy: {kinetic_energy}")
    print(f"Total Final Energy: {e_energy + b_energy + kinetic_energy}\n")


    duration = end - start
    # calculate the total simulation time

    simulation_stats = {
        "total_time": duration,
        "total_iterations": Nt,
        "time_per_iteration": duration / Nt
    }

    dump_parameters_to_toml(simulation_stats, simulation_parameters, plasma_parameters, plotting_parameters, constants, particles)
    # save the parameters to an output file

    print(f"\nSimulation Complete")
    print(f"Total Simulation Time: {duration} s")
    print(f"Time Per Iteration: {duration/Nt} s")


if __name__ == "__main__":
    main()
    # run the main function


from jax import jit
from jax.tree_util import register_pytree_node_class
import jax.numpy as jnp

# Christopher Woolford, Oct 22 2024
# This file contains functions that apply boundary conditions to a field.

def apply_supergaussian_boundary_condition(field, boundary_thickness, order, strength):
    """
    Apply Super-Gaussian absorbing boundary conditions to the given field.

    Args:
        field (ndarray): The field to which the Super-Gaussian boundary condition is applied.
        boundary_thickness (int): The thickness of the absorbing boundary layer.
        order (int): The order of the Super-Gaussian function.
        strength (float): The strength of the absorption.

    Returns:
        ndarray: The field with Super-Gaussian boundary conditions applied.
    """
    def supergaussian_factor(x, thickness, order, strength):
        return jnp.exp(-strength * (x / thickness)**order)

    nx, ny, nz = field.shape
    for i in range(boundary_thickness):
        factor = supergaussian_factor(i, boundary_thickness, order, strength)
        field = field.at[i, :, :].mul(factor)
        field = field.at[nx - 1 - i, :, :].mul(factor)
        field = field.at[:, i, :].mul(factor)
        field = field.at[:, ny - 1 - i, :].mul(factor)
        field = field.at[:, :, i].mul(factor)
        field = field.at[:, :, nz - 1 - i].mul(factor)

    return field

@jit
def apply_zero_boundary_condition(field):
    """
    Apply zero boundary conditions to the given field.

    Args:
        field (ndarray): The field to which the zero boundary condition is applied.

    Returns:
        ndarray: The field with zero boundary conditions applied.
    """
    field = field.at[0, :, :].set(0)
    field = field.at[-1, :, :].set(0)
    field = field.at[:, 0, :].set(0)
    field = field.at[:, -1, :].set(0)
    field = field.at[:, :, 0].set(0)
    field = field.at[:, :, -1].set(0)

    return field


import time
import jax
from jax import random
import jax.numpy as jnp
import os, sys
import matplotlib.pyplot as plt
# import external libraries

jax.config.update('jax_platform_name', 'cpu')

from . import errors
from . import fields
from . import boundaryconditions
from . import initialization
from . import particle
from . import plotting
from . import utils
from . import pstd
from . import fdtd
from . import boris
from . import rho
from . import evolve


from jax import jit
import jax.numpy as jnp
from functools import partial
# import external libraries

from PyPIC3D.pstd import spectral_laplacian, spectral_divergence
from PyPIC3D.fdtd import centered_finite_difference_laplacian, centered_finite_difference_divergence
# import functions from the PyPIC3D package

@partial(jit, static_argnums=(4, 5))
def compute_pe(phi, rho, constants, world, solver, bc='periodic'):
    """
    Compute the relative percentage difference of the Poisson solver.

    Args:
        phi (ndarray): The potential field.
        rho (ndarray): The charge density.
        eps (float): The permittivity.
        dx (float): The grid spacing in the x-direction.
        dy (float): The grid spacing in the y-direction.
        dz (float): The grid spacing in the z-direction.
        bc (str): The boundary condition.

    Returns:
        float: The relative percentage difference of the Poisson solver.
    """
    eps = constants['eps']
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    if solver == 'spectral':
        x = spectral_laplacian(phi, world)
    elif solver == 'fdtd':
        x = centered_finite_difference_laplacian(phi, dx, dy, dz, bc)
    elif solver == 'autodiff':
        return 0
    poisson_error = x + rho/eps
    magnitude = jnp.mean(jnp.abs(rho/eps)) + 1e-16
    return jnp.mean(jnp.abs(poisson_error)) / magnitude

@partial(jit, static_argnums=(4, 5))
def compute_magnetic_divergence_error(Bx, By, Bz, world, solver, bc='periodic'):
    """
    Compute the error in the divergence of the magnetic field for different solvers.

    Args:
        Bx (ndarray): The x-component of the magnetic field.
        By (ndarray): The y-component of the magnetic field.
        Bz (ndarray): The z-component of the magnetic field.
        dx (float): The grid spacing in the x-direction.
        dy (float): The grid spacing in the y-direction.
        dz (float): The grid spacing in the z-direction.
        bc (str): The boundary condition.

    Returns:
        float: The error in the divergence of the magnetic field.
    """
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']

    if solver == 'spectral':
        divB = spectral_divergence(Bx, By, Bz, world)
    elif solver == 'fdtd':
        divB = centered_finite_difference_divergence(Bx, By, Bz, dx, dy, dz, bc)
    elif solver == 'autodiff':
        return 0
    divergence_error = jnp.mean(jnp.abs(divB))
    return divergence_error

@partial(jit, static_argnums=(6, 7))
def compute_electric_divergence_error(Ex, Ey, Ez, rho, constants, world, solver, bc='periodic'):
    """
    Compute the error in the divergence of the electric field using the charge density and the components of the electric field.

    Args:
        Ex (ndarray): The x-component of the electric field.
        Ey (ndarray): The y-component of the electric field.
        Ez (ndarray): The z-component of the electric field.
        rho (ndarray): The charge density.
        eps (float): The permittivity.
        dx (float): The grid spacing in the x-direction.
        dy (float): The grid spacing in the y-direction.
        dz (float): The grid spacing in the z-direction.
        bc (str): The boundary condition.

    Returns:
    float: The error in the divergence of the electric field.
    """
    eps = constants['eps']
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']

    if solver == 'spectral':
        divE = spectral_divergence(Ex, Ey, Ez, world)
    elif solver == 'fdtd':
        divE = centered_finite_difference_divergence(Ex, Ey, Ez, dx, dy, dz, bc)
    elif solver == 'autodiff':
        return 0

    divergence_error = jnp.mean(jnp.abs(divE - rho / eps))
    return divergence_error


from jax import jit
import jax
import jax.numpy as jnp
from functools import partial
# import external libraries

def create_k_mesh(nx, ny, nz, dx, dy, dz):
    """
    Create a mesh of wave numbers for FFT operations.

    Args:
        nx (int): Number of grid points in the x-direction.
        ny (int): Number of grid points in the y-direction.
        nz (int): Number of grid points in the z-direction.
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.

    Returns:
        tuple: Meshgrid of wave numbers (kx, ky, kz).
    """
    kx = jnp.fft.fftfreq(nx, d=dx) * 2 * jnp.pi
    ky = jnp.fft.fftfreq(ny, d=dy) * 2 * jnp.pi
    kz = jnp.fft.fftfreq(nz, d=dz) * 2 * jnp.pi
    return jnp.meshgrid(kx, ky, kz, indexing='ij')

@jit
def spectral_poisson_solve(rho, constants, world):
    """
    Solve the Poisson equation for electrostatic potential using spectral method.

    Args:
        rho (ndarray): Charge density.
        eps (float): Permittivity.
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.

    Returns:
        phi (ndarray): Solution to the Poisson equation.
    """
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    eps = constants['eps']

    krho = jnp.fft.fftn(rho)
    nx, ny, nz = krho.shape
    kx, ky, kz = create_k_mesh(nx, ny, nz, dx, dy, dz)
    k2 = kx**2 + ky**2 + kz**2
    k2 = k2.at[0, 0, 0].set(1.0)
    phi = krho / (eps*k2)
    phi = phi.at[0, 0, 0].set(0)
    phi = jnp.fft.ifftn(phi).real
    return phi

@jit
def spectral_divergence(xfield, yfield, zfield, world):
    """
    Calculate the spectral divergence of a 3D vector field.

    This function computes the divergence of a vector field in the spectral domain
    using the Fast Fourier Transform (FFT). The input fields are assumed to be 
    periodic in all three dimensions.

    Args:
        xfield (ndarray): The x-component of the vector field.
        yfield (ndarray): The y-component of the vector field.
        zfield (ndarray): The z-component of the vector field.
        dx (float): The grid spacing in the x-direction.
        dy (float): The grid spacing in the y-direction.
        dz (float): The grid spacing in the z-direction.

    Returns:
        ndarray: The real part of the inverse FFT of the spectral divergence.
    """

    xfft = jnp.fft.fftn(xfield)
    yfft = jnp.fft.fftn(yfield)
    zfft = jnp.fft.fftn(zfield)
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    nx, ny, nz = xfft.shape
    kx, ky, kz = create_k_mesh(nx, ny, nz, dx, dy, dz)
    div = 1j * kx * xfft + 1j * ky * yfft + 1j * kz * zfft
    return jnp.fft.ifftn(div).real

@jit
def spectral_curl(xfield, yfield, zfield, world):
    """
    Compute the curl of a 3D vector field using spectral methods.

    Args:
        xfield (ndarray): The x-component of the vector field.
        yfield (ndarray): The y-component of the vector field.
        zfield (ndarray): The z-component of the vector field.
        dx (float): The grid spacing in the x-direction.
        dy (float): The grid spacing in the y-direction.
        dz (float): The grid spacing in the z-direction.

    Returns:
        tuple: A tuple containing the x, y, and z components of the curl of the vector field.
    """

    xfft = jnp.fft.fftn(xfield)
    yfft = jnp.fft.fftn(yfield)
    zfft = jnp.fft.fftn(zfield)
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    nx, ny, nz = xfft.shape
    kx, ky, kz = create_k_mesh(nx, ny, nz, dx, dy, dz)
    curlx = 1j * ky * zfft - 1j * kz * yfft
    curly = 1j * kz * xfft - 1j * kx * zfft
    curlz = 1j * kx * yfft - 1j * ky * xfft
    return jnp.fft.ifftn(curlx).real, jnp.fft.ifftn(curly).real, jnp.fft.ifftn(curlz).real

@jit
def spectral_gradient(field, world):
    """
    Compute the gradient of a 3D scalar field using spectral methods.

    Args:
        field (ndarray): The input scalar field.
        dx (float): The grid spacing in the x-direction.
        dy (float): The grid spacing in the y-direction.
        dz (float): The grid spacing in the z-direction.

    Returns:
        tuple: A tuple containing the x, y, and z components of the gradient of the field.
    """

    field_fft = jnp.fft.fftn(field)
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    nx, ny, nz = field_fft.shape
    kx, ky, kz = create_k_mesh(nx, ny, nz, dx, dy, dz)
    gradx = 1j * kx * field_fft
    grady = 1j * ky * field_fft
    gradz = 1j * kz * field_fft
    return jnp.fft.ifftn(gradx).real, jnp.fft.ifftn(grady).real, jnp.fft.ifftn(gradz).real

@jit
def spectral_laplacian(field, world):
    """
    Calculates the Laplacian of a given field using spectral method. Assumes periodic boundary conditions.

    Args:
        field: numpy.ndarray
            The input field.
        dx: float
            The spacing between grid points in the x-direction.
        dy: float
            The spacing between grid points in the y-direction.
        dz: float
            The spacing between grid points in the z-direction.

    Returns:
        numpy.ndarray
            The Laplacian of the field.
    """
    field_fft = jnp.fft.fftn(field)
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    nx, ny, nz = field_fft.shape
    kx, ky, kz = create_k_mesh(nx, ny, nz, dx, dy, dz)
    lapl = -(kx**2 + ky**2 + kz**2) * field_fft
    return jnp.fft.ifftn(lapl).real


import jax
from jax import jit
import jax.numpy as jnp
from functools import partial
from jax import lax
# import external libraries

from PyPIC3D.rho import compute_rho

@jit
def J_from_rhov(particles, J, constants, world, grid):
    """
    Compute the current density from the charge density and particle velocities.

    Args:
        particles (list): List of particle species, each with methods to get charge, subcell position, resolution, and index.
        rho (ndarray): Charge density array.
        J (tuple): Current density arrays (Jx, Jy, Jz) for the x, y, and z directions respectively.
        constants (dict): Dictionary containing physical constants.

    Returns:
        tuple: Updated current density arrays (Jx, Jy, Jz) for the x, y, and z directions respectively.
    """

    C = constants['C']
    # speed of light

    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    # get the world parameters

    Jx, Jy, Jz = J
    # unpack the values of J
    Jx = Jx.at[:, :, :].set(0)
    Jy = Jy.at[:, :, :].set(0)
    Jz = Jz.at[:, :, :].set(0)
    # initialize the current arrays as 0
    J = (Jx, Jy, Jz)
    # initialize the current density as a tuple

    for species in particles:
        N_particles = species.get_number_of_particles()
        charge = species.get_charge()
        # get the number of particles and their charge
        particle_x, particle_y, particle_z = species.get_position()
        # get the position of the particles in the species

        vx, vy, vz = species.get_velocity()
        # get the velocity of the particles in the species

        shape_factor = species.get_shape()

        def add_to_J(particle, J):
            x = particle_x.at[particle].get()
            y = particle_y.at[particle].get()
            z = particle_z.at[particle].get()
            # get the position of the particle
            vx_particle = vx.at[particle].get()
            vy_particle = vy.at[particle].get()
            vz_particle = vz.at[particle].get()
            # get the velocity of the particle

            J = lax.cond(
                shape_factor == 1,
                lambda _: J_first_order_weighting(charge, x, y, z, vx_particle, vy_particle, vz_particle, J, dx, dy, dz,  grid),
                lambda _: J_second_order_weighting(charge, x, y, z, vx_particle, vy_particle, vz_particle, J, dx, dy, dz, grid),
                operand=None
            )
            # use first-order weighting to distribute the current density

            return J
        # add the particle species to the charge density array
        J = jax.lax.fori_loop(0, N_particles, add_to_J, J)
    # loop over all particles in the species and add their contribution to the current density

    return J


@jit
def J_first_order_weighting(q, x, y, z, vx, vy, vz, J, dx, dy, dz, grid):

    Jx, Jy, Jz = J
    # unpack the values of J

    Nx, Ny, Nz = Jx.shape
    # get the shape of the charge density array

    x0 = jnp.floor((x - grid[0][0]) / dx).astype(int)
    y0 = jnp.floor((y - grid[1][0]) / dy).astype(int)
    z0 = jnp.floor((z - grid[2][0]) / dz).astype(int)
    # Calculate the nearest grid points

    deltax = x - jnp.floor(x / dx) * dx
    deltay = y - jnp.floor(y / dy) * dy
    deltaz = z - jnp.floor(z / dz) * dz
    # Calculate the difference between the particle position and the nearest grid point


    x1 = wrap_around(x0 + 1, Nx)
    y1 = wrap_around(y0 + 1, Ny)
    z1 = wrap_around(z0 + 1, Nz)
    # Calculate the index of the next grid point

    dv = dx * dy * dz
    # Calculate the volume of each grid point

    ##################################### WEIGHTING FACTORS ##########################################################
    Sx0 = 1 - (deltax/dx)
    Sy0 = 1 - (deltay/dy)
    Sz0 = 1 - (deltaz/dz)
    # calculate the weight for the center grid points
    Sx1 = deltax/dx
    Sy1 = deltay/dy
    Sz1 = deltaz/dz
    # calculate the weight for the next grid points
    ####################################################################################################################

    ####################################### JX DISTRIBUTION ########################################################
    Jx = Jx.at[x0, y0, z0].add((q * vx / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    Jx = Jx.at[x1, y0, z0].add((q * vx / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    Jx = Jx.at[x0, y1, z0].add((q * vx / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    Jx = Jx.at[x0, y0, z1].add((q * vx / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    Jx = Jx.at[x1, y1, z0].add((q * vx / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    Jx = Jx.at[x1, y0, z1].add((q * vx / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    Jx = Jx.at[x0, y1, z1].add((q * vx / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    Jx = Jx.at[x1, y1, z1].add((q * vx / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    # Distribute the current density in the x direction to the grid points
    #####################################################################################################################

    ####################################### JY DISTRIBUTION ########################################################
    Jy = Jy.at[x0, y0, z0].add((q * vy / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    Jy = Jy.at[x1, y0, z0].add((q * vy / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    Jy = Jy.at[x0, y1, z0].add((q * vy / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    Jy = Jy.at[x0, y0, z1].add((q * vy / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    Jy = Jy.at[x1, y1, z0].add((q * vy / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    Jy = Jy.at[x1, y0, z1].add((q * vy / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    Jy = Jy.at[x0, y1, z1].add((q * vy / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    Jy = Jy.at[x1, y1, z1].add((q * vy / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    # Distribute the current density in the y direction to the grid points
    #####################################################################################################################

    ####################################### JZ DISTRIBUTION ########################################################
    Jz = Jz.at[x0, y0, z0].add((q * vz / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    Jz = Jz.at[x1, y0, z0].add((q * vz / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    Jz = Jz.at[x0, y1, z0].add((q * vz / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    Jz = Jz.at[x0, y0, z1].add((q * vz / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    Jz = Jz.at[x1, y1, z0].add((q * vz / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    Jz = Jz.at[x1, y0, z1].add((q * vz / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    Jz = Jz.at[x0, y1, z1].add((q * vz / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    Jz = Jz.at[x1, y1, z1].add((q * vz / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    # Distribute the current density in the z direction to the grid points
    #####################################################################################################################


    return (Jx, Jy, Jz)


@jit
def J_second_order_weighting(q, x, y, z, vx, vy, vz, J, dx, dy, dz, grid):
    """
    Distribute the current of a particle to the surrounding grid points using second-order weighting.

    Args:
        q (float): Charge of the particle.
        x, y, z (float): Position of the particle.
        vx, vy, vz (float): Velocity components of the particle.
        J (tuple): Current density arrays (Jx, Jy, Jz).
        rho (ndarray): Charge density array (for shape).
        dx, dy, dz (float): Grid spacing.
        x_wind, y_wind, z_wind (float): Window offsets.

    Returns:
        tuple: Updated current density arrays (Jx, Jy, Jz).
    """

    Jx, Jy, Jz = J
    Nx, Ny, Nz = Jx.shape

    x0 = jnp.floor((x - grid[0][0]) / dx).astype(int)
    y0 = jnp.floor((y - grid[1][0]) / dy).astype(int)
    z0 = jnp.floor((z - grid[2][0]) / dz).astype(int)
    # Calculate the nearest grid points

    deltax = x - jnp.floor( x / dx) * dx
    deltay = y - jnp.floor( y / dy) * dy
    deltaz = z - jnp.floor( z / dz) * dz
    # Calculate the difference between the particle position and the nearest grid point

    x1 = wrap_around(x0 + 1, Nx)
    y1 = wrap_around(y0 + 1, Ny)
    z1 = wrap_around(z0 + 1, Nz)

    x_minus1 = x0 - 1
    y_minus1 = y0 - 1
    z_minus1 = z0 - 1

    dv = dx * dy * dz

    # Weighting factors
    Sx0 = (3/4) - (deltax/dx)**2
    Sy0 = (3/4) - (deltay/dy)**2
    Sz0 = (3/4) - (deltaz/dz)**2

    Sx1 = jax.lax.cond(
        deltax <= dx/2,
        lambda _: (1/2) * ((1/2) - (deltax/dx))**2,
        lambda _: jnp.array(0.0, dtype=deltax.dtype),
        operand=None
    )
    Sy1 = jax.lax.cond(
        deltay <= dy/2,
        lambda _: (1/2) * ((1/2) - (deltay/dy))**2,
        lambda _: jnp.array(0.0, dtype=deltay.dtype),
        operand=None
    )
    Sz1 = jax.lax.cond(
        deltaz <= dz/2,
        lambda _: (1/2) * ((1/2) - (deltaz/dz))**2,
        lambda _: jnp.array(0.0, dtype=deltaz.dtype),
        operand=None
    )

    Sx_minus1 = jax.lax.cond(
        deltax <= dx/2,
        lambda _: (1/2) * ((1/2) + (deltax/dx))**2,
        lambda _: jnp.array(0.0, dtype=deltax.dtype),
        operand=None
    )
    Sy_minus1 = jax.lax.cond(
        deltay <= dy/2,
        lambda _: (1/2) * ((1/2) + (deltay/dy))**2,
        lambda _: jnp.array(0.0, dtype=deltay.dtype),
        operand=None
    )
    Sz_minus1 = jax.lax.cond(
        deltaz <= dz/2,
        lambda _: (1/2) * ((1/2) + (deltaz/dz))**2,
        lambda _: jnp.array(0.0, dtype=deltaz.dtype),
        operand=None
    )

    # Jx distribution
    Jx = Jx.at[x0, y0, z0].add((q * vx / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    Jx = Jx.at[x1, y0, z0].add((q * vx / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    Jx = Jx.at[x0, y1, z0].add((q * vx / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    Jx = Jx.at[x0, y0, z1].add((q * vx / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    Jx = Jx.at[x1, y1, z0].add((q * vx / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    Jx = Jx.at[x1, y0, z1].add((q * vx / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    Jx = Jx.at[x0, y1, z1].add((q * vx / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    Jx = Jx.at[x1, y1, z1].add((q * vx / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    Jx = Jx.at[x_minus1, y0, z0].add((q * vx / dv) * Sx_minus1 * Sy0 * Sz0, mode='drop')
    Jx = Jx.at[x0, y_minus1, z0].add((q * vx / dv) * Sx0 * Sy_minus1 * Sz0, mode='drop')
    Jx = Jx.at[x0, y0, z_minus1].add((q * vx / dv) * Sx0 * Sy0 * Sz_minus1, mode='drop')
    Jx = Jx.at[x_minus1, y_minus1, z0].add((q * vx / dv) * Sx_minus1 * Sy_minus1 * Sz0, mode='drop')
    Jx = Jx.at[x_minus1, y0, z_minus1].add((q * vx / dv) * Sx_minus1 * Sy0 * Sz_minus1, mode='drop')
    Jx = Jx.at[x0, y_minus1, z_minus1].add((q * vx / dv) * Sx0 * Sy_minus1 * Sz_minus1, mode='drop')
    Jx = Jx.at[x_minus1, y_minus1, z_minus1].add((q * vx / dv) * Sx_minus1 * Sy_minus1 * Sz_minus1, mode='drop')

    # Jy distribution
    Jy = Jy.at[x0, y0, z0].add((q * vy / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    Jy = Jy.at[x1, y0, z0].add((q * vy / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    Jy = Jy.at[x0, y1, z0].add((q * vy / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    Jy = Jy.at[x0, y0, z1].add((q * vy / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    Jy = Jy.at[x1, y1, z0].add((q * vy / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    Jy = Jy.at[x1, y0, z1].add((q * vy / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    Jy = Jy.at[x0, y1, z1].add((q * vy / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    Jy = Jy.at[x1, y1, z1].add((q * vy / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    Jy = Jy.at[x_minus1, y0, z0].add((q * vy / dv) * Sx_minus1 * Sy0 * Sz0, mode='drop')
    Jy = Jy.at[x0, y_minus1, z0].add((q * vy / dv) * Sx0 * Sy_minus1 * Sz0, mode='drop')
    Jy = Jy.at[x0, y0, z_minus1].add((q * vy / dv) * Sx0 * Sy0 * Sz_minus1, mode='drop')
    Jy = Jy.at[x_minus1, y_minus1, z0].add((q * vy / dv) * Sx_minus1 * Sy_minus1 * Sz0, mode='drop')
    Jy = Jy.at[x_minus1, y0, z_minus1].add((q * vy / dv) * Sx_minus1 * Sy0 * Sz_minus1, mode='drop')
    Jy = Jy.at[x0, y_minus1, z_minus1].add((q * vy / dv) * Sx0 * Sy_minus1 * Sz_minus1, mode='drop')
    Jy = Jy.at[x_minus1, y_minus1, z_minus1].add((q * vy / dv) * Sx_minus1 * Sy_minus1 * Sz_minus1, mode='drop')

    # Jz distribution
    Jz = Jz.at[x0, y0, z0].add((q * vz / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    Jz = Jz.at[x1, y0, z0].add((q * vz / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    Jz = Jz.at[x0, y1, z0].add((q * vz / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    Jz = Jz.at[x0, y0, z1].add((q * vz / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    Jz = Jz.at[x1, y1, z0].add((q * vz / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    Jz = Jz.at[x1, y0, z1].add((q * vz / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    Jz = Jz.at[x0, y1, z1].add((q * vz / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    Jz = Jz.at[x1, y1, z1].add((q * vz / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    Jz = Jz.at[x_minus1, y0, z0].add((q * vz / dv) * Sx_minus1 * Sy0 * Sz0, mode='drop')
    Jz = Jz.at[x0, y_minus1, z0].add((q * vz / dv) * Sx0 * Sy_minus1 * Sz0, mode='drop')
    Jz = Jz.at[x0, y0, z_minus1].add((q * vz / dv) * Sx0 * Sy0 * Sz_minus1, mode='drop')
    Jz = Jz.at[x_minus1, y_minus1, z0].add((q * vz / dv) * Sx_minus1 * Sy_minus1 * Sz0, mode='drop')
    Jz = Jz.at[x_minus1, y0, z_minus1].add((q * vz / dv) * Sx_minus1 * Sy0 * Sz_minus1, mode='drop')
    Jz = Jz.at[x0, y_minus1, z_minus1].add((q * vz / dv) * Sx0 * Sy_minus1 * Sz_minus1, mode='drop')
    Jz = Jz.at[x_minus1, y_minus1, z_minus1].add((q * vz / dv) * Sx_minus1 * Sy_minus1 * Sz_minus1, mode='drop')

    return (Jx, Jy, Jz)

@jit
def wrap_around(ix, size):
    """Wrap around index (scalar or 1D array) to ensure it is within bounds."""
    return jnp.where(ix > size - 1, ix - size, ix)

@jit
def VB_current(particles, J, constants, world, grid):
    """
    Apply Villasenor-Buneman correction to ensure rigorous charge conservation for local electromagnetic field solvers.

    Args:
        particles (list): List of particle species, each with methods to get charge, subcell position, resolution, and index.
        Nx (int): Number of grid points in the x-direction.
        Ny (int): Number of grid points in the y-direction.
        Nz (int): Number of grid points in the z-direction.

    Returns:
        tuple: Corrected current density arrays (Jx, Jy, Jz) for the x, y, and z directions respectively.

    References:
        Villasenor, J., & Buneman, O. (1992). Rigorous charge conservation for local electromagnetic field solvers.
        Computer Physics Communications, 69(2-3), 306-316.
    """

    Jx, Jy, Jz = J
    # unpack the values of J

    C = constants['C']
    # speed of light

    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']
    # get the world parameters

    for species in particles:
        q = species.get_charge()
        # get the charge of the species
        old_x, old_y, old_z = species.get_old_position()
        # get the old position of the particles in the species
        x, y, z = species.get_position()
        # get the position of the particles in the species

        zeta1 = jnp.floor((x + x_wind / 2) / dx).astype(int)
        eta1 = jnp.floor((y + y_wind / 2) / dy).astype(int)
        xi1 = jnp.floor((z + z_wind / 2) / dz).astype(int)
        zeta2 = jnp.floor((old_x + x_wind / 2) / dx).astype(int)
        eta2 = jnp.floor((old_y + y_wind / 2) / dy).astype(int)
        xi2 = jnp.floor((old_z + z_wind / 2) / dz).astype(int)
        # Calculate the nearest grid points

        dx, dy, dz = species.get_resolution()
        # get the resolution of the species

        deltax = zeta2 - zeta1
        deltay = eta2 - eta1
        deltaz = xi2 - xi1

        zetabar = 0.5*(zeta1 + zeta2)
        etabar = 0.5*(eta1 + eta2)
        xibar = 0.5*(xi1 + xi2)
        # compute the displacement variables from Buneman, Vilasenor 1991

        ix, iy, iz = species.get_index()
        # get the index of the species

        dq = q/dx/dy/dz
        # charge differential

        ix1 = wrap_around(ix + 1, Jx.shape[0])
        iy1 = wrap_around(iy + 1, Jx.shape[1])
        iz1 = wrap_around(iz + 1, Jx.shape[2])

        Jx = Jx.at[ix, iy1, iz1].add( dq*C*(deltax*zetabar*xibar + deltax*deltay*deltaz/12))
        # compute the first x correction for charge conservation along i+1/2, j, k
        Jx = Jx.at[ix, iy, iz1].add( dq*C*(deltax*(1-etabar)*xibar - deltax*deltay*deltaz/12))
        # compute the second x correction for charge conservation along i+1/2, j, k+1
        Jx = Jx.at[ix, iy1, iz].add( dq*C*(deltax*etabar*(1-xibar) - deltax*deltay*deltaz/12))
        # compute the third x correction for charge conservation along i+1/2, j+1, k
        Jx = Jx.at[ix, iy, iz].add( dq*C*(deltax*(1-etabar)*(1-xibar) + deltax*deltay*deltaz/12))
        # compute the fourth x correction for charge conservation along i+1/2, j, k

        Jy = Jy.at[ix1, iy, iz1].add( dq*C*(deltay*zetabar*xibar + deltax*deltay*deltaz/12))
        # compute the first y correction for charge conservation along i, j+1/2, k
        Jy = Jy.at[ix1, iy, iz].add( dq*C*(deltay*(1-zetabar)*xibar - deltax*deltay*deltaz/12))
        # compute the second y correction for charge conservation along i+1, j+1/2, k
        Jy = Jy.at[ix, iy, iz1].add( dq*C*(deltay*zetabar*(1-xibar) - deltax*deltay*deltaz/12))
        # compute the third y correction for charge conservation along i, j+1/2, k+1
        Jy = Jy.at[ix, iy, iz].add( dq*C*(deltay*(1-zetabar)*(1-xibar) + deltax*deltay*deltaz/12))
        # compute the fourth y correction for charge conservation along i, j+1/2, k

        Jz = Jz.at[ix1, iy1, iz].add(dq*C*(deltaz*(1-zetabar)*etabar + deltax*deltay*deltaz/12))
        # compute the first z correction for charge conservation along i+1, j+1, k+1/2
        Jz = Jz.at[ix, iy1, iz].add(dq*C*(deltaz*(1-zetabar)*(1-etabar) - deltax*deltay*deltaz/12))
        # compute the second z correction for charge conservation along i, j+1, k+1/2
        Jz = Jz.at[ix1, iy, iz].add(dq*C*(deltaz*zetabar*(1-etabar) - deltax*deltay*deltaz/12))
        # compute the third z correction for charge conservation along i+1, j, k+1/2
        Jz = Jz.at[ix, iy, iz].add(dq*C*(deltaz*(1-zetabar)*(1-etabar) + deltax*deltay*deltaz/12))
        # compute the fourth z correction for charge conservation along i, j, k+1/2

    return (Jx, Jy, Jz)
    # return the current corrections

def Esirkepov_current(particles, J, constants, world, grid):

    Jx, Jy, Jz = J
    # unpack the values of J

    Jx = Jx.at[:, :, :].set(0)
    Jy = Jy.at[:, :, :].set(0)
    Jz = Jz.at[:, :, :].set(0)
    # initialize the current arrays as 0

    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    # get the grid spacing parameters
    dt = world['dt']
    # get the time step
    Nx = world['Nx']
    Ny = world['Ny']
    Nz = world['Nz']
    # get the number of grid points in each direction


    for species in particles:
        q = species.get_charge()
        # get the charge of the species

        old_x, old_y, old_z = species.get_old_position()
        # get the old position of the particles in the species
        x, y, z = species.get_position()
        # get the position of the particles in the species

        shape_factor = species.get_shape()
        # get the shape factor of the species

        x0 = jnp.floor((x - grid[0][0]) / dx).astype(int)
        y0 = jnp.floor((y - grid[1][0]) / dy).astype(int)
        z0 = jnp.floor((z - grid[2][0]) / dz).astype(int)
        # Calculate the nearest grid points

        x1 = wrap_around(x0 + 1, Nx)
        y1 = wrap_around(y0 + 1, Ny)
        z1 = wrap_around(z0 + 1, Nz)
        # Calculate the index of the next grid point

        x_minus1 = wrap_around(x1 - 1, Nx)
        y_minus1 = wrap_around(y1 - 1, Ny)
        z_minus1 = wrap_around(z1 - 1, Nz)
        # Calculate the index of the next grid point

        xpts = [x_minus1, x0, x1]
        ypts = [y_minus1, y0, y1]
        zpts = [z_minus1, z0, z1]
        # List of grid points in each direction

        old_deltax = old_x - jnp.floor(old_x / dx) * dx
        old_deltay = old_y - jnp.floor(old_y / dy) * dy
        old_deltaz = old_z - jnp.floor(old_z / dz) * dz
        # Calculate the difference between the particle position and the nearest grid point
        new_deltax = x - jnp.floor(x / dx) * dx
        new_deltay = y - jnp.floor(y / dy) * dy
        new_deltaz = z - jnp.floor(z / dz) * dz
        # Calculate the difference between the particle position and the nearest grid point

        old_x_weights, old_y_weights, old_z_weights = jax.lax.cond(
            shape_factor == 1,
            lambda _: get_first_order_weights(old_deltax, old_deltay, old_deltaz, dx, dy, dz),
            lambda _: get_second_order_weights(old_deltax, old_deltay, old_deltaz, dx, dy, dz),
            operand=None
        )
        x_weights, y_weights, z_weights = jax.lax.cond(
            shape_factor == 1,
            lambda _: get_first_order_weights(new_deltax, new_deltay, new_deltaz, dx, dy, dz),
            lambda _: get_second_order_weights(new_deltax, new_deltay, new_deltaz, dx, dy, dz),
            operand=None
        )
        # Calculate the weights for the grid points

        d_Sx = []
        d_Sy = []
        d_Sz = []

        for i in range(len(x_weights)):
            d_Sx.append(x_weights[i] - old_x_weights[i])
            d_Sy.append(y_weights[i] - old_y_weights[i])
            d_Sz.append(z_weights[i] - old_z_weights[i])
        # Calculate the difference in weights for the central grid points

        Wx = []
        Wy = []
        Wz = []

        for i in range(len(x_weights)):
            Wx.append( d_Sx[i] * ( old_y_weights[i] * old_z_weights[i] + 0.5 * d_Sy[i] * old_z_weights[i] \
               + 0.5 * old_y_weights[i] * d_Sz[i] + d_Sy[i] * d_Sz[i] / 3))
            Wy.append( d_Sy[i] * ( old_x_weights[i] * old_z_weights[i] + 0.5 * d_Sx[i] * old_z_weights[i] \
               + 0.5 * old_x_weights[i] * d_Sz[i] + d_Sx[i] * d_Sz[i] / 3))
            Wz.append( d_Sz[i] * ( old_x_weights[i] * old_y_weights[i] + 0.5 * d_Sx[i] * old_y_weights[i] \
               + 0.5 * old_x_weights[i] * d_Sy[i] + d_Sx[i] * d_Sy[i] / 3))


        for i in range(len(xpts)):
            x = xpts[i]
            y = ypts[i]
            z = zpts[i]
            # Loop over the grid points in each direction
            Jx = Jx.at[x, y, z].add( Wx[i] *  (q / dy / dz / dt), mode='drop')
            Jy = Jy.at[x, y, z].add( Wy[i] *  (q / dx / dz / dt), mode='drop')
            Jz = Jz.at[x, y, z].add( Wz[i] *  (q / dx / dy / dt), mode='drop')
            # Distribute the current density to the grid points

    return (Jx, Jy, Jz)

def get_second_order_weights(deltax, deltay, deltaz, dx, dy, dz):
    """
    Calculate the second-order weights for particle current distribution.

    Args:
        deltax, deltay, deltaz (float): Particle position offsets from grid points.
        dx, dy, dz (float): Grid spacings in x, y, and z directions.

    Returns:
        tuple: Weights for x, y, and z directions.
    """
    Sx0 = (3/4) - (deltax/dx)**2
    Sy0 = (3/4) - (deltay/dy)**2
    Sz0 = (3/4) - (deltaz/dz)**2

    Sx1 = jnp.where(deltax <= dx/2, (1/2) * ((1/2) - (deltax/dx))**2, 0.0)
    Sy1 = jnp.where(deltay <= dy/2, (1/2) * ((1/2) - (deltay/dy))**2, 0.0)
    Sz1 = jnp.where(deltaz <= dz/2, (1/2) * ((1/2) - (deltaz/dz))**2, 0.0)

    Sx_minus1 = jnp.where(deltax <= dx/2, (1/2) * ((1/2) + (deltax/dx))**2, 0.0)
    Sy_minus1 = jnp.where(deltay <= dy/2, (1/2) * ((1/2) + (deltay/dy))**2, 0.0)
    Sz_minus1 = jnp.where(deltaz <= dz/2, (1/2) * ((1/2) + (deltaz/dz))**2, 0.0)

    x_weights = [Sx_minus1, Sx0, Sx1]
    y_weights = [Sy_minus1, Sy0, Sy1]
    z_weights = [Sz_minus1, Sz0, Sz1]

    return x_weights, y_weights, z_weights


def get_first_order_weights(deltax, deltay, deltaz, dx, dy, dz):
    """
    Calculate the first-order weights for particle current distribution.

    Args:
        deltax, deltay, deltaz (float): Particle position offsets from grid points.
        dx, dy, dz (float): Grid spacings in x, y, and z directions.

    Returns:
        tuple: Weights for x, y, and z directions.
    """
    Sx0 = jnp.asarray(1 - deltax / dx)
    Sy0 = jnp.asarray(1 - deltay / dy)
    Sz0 = jnp.asarray(1 - deltaz / dz)

    Sx1 = jnp.asarray(deltax / dx)
    Sy1 = jnp.asarray(deltay / dy)
    Sz1 = jnp.asarray(deltaz / dz)

    Sx_minus1 = jnp.zeros_like(Sx0)
    Sy_minus1 = jnp.zeros_like(Sy0)
    Sz_minus1 = jnp.zeros_like(Sz0)
    # No second-order weights for first-order weighting

    x_weights = [Sx_minus1, Sx0, Sx1]
    y_weights = [Sy_minus1, Sy0, Sy1]
    z_weights = [Sz_minus1, Sz0, Sz1]

    return x_weights, y_weights, z_weights


import jax
from jax import jit
import jax.numpy as jnp
from jax import lax
# import external libraries

@jit
def compute_rho(particles, rho, world):
    """
    Compute the charge density (rho) for a given set of particles in a simulation world.
    Parameters:
    particles (list): A list of particle species, each containing methods to get the number of particles,
                      their positions, and their charge.
    rho (ndarray): The initial charge density array to be updated.
    world (dict): A dictionary containing the simulation world parameters, including:
                  - 'dx': Grid spacing in the x-direction.
                  - 'dy': Grid spacing in the y-direction.
                  - 'dz': Grid spacing in the z-direction.
                  - 'x_wind': Window size in the x-direction.
                  - 'y_wind': Window size in the y-direction.
                  - 'z_wind': Window size in the z-direction.
    Returns:
    ndarray: The updated charge density array.
    """
    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    x_wind = world['x_wind']
    y_wind = world['y_wind']
    z_wind = world['z_wind']

    rho = jnp.zeros_like(rho)
    # reset rho to zero

    for species in particles:
        shape_factor = species.get_shape()
        # get the shape factor of the species, which determines the weighting function
        N_particles = species.get_number_of_particles()
        charge = species.get_charge()
        # get the number of particles and their charge
        particle_x, particle_y, particle_z = species.get_position()
        # get the position of the particles in the species

        def add_to_rho(particle, rho):
            x = particle_x.at[particle].get()
            y = particle_y.at[particle].get()
            z = particle_z.at[particle].get()
            # get the position of the particle
            rho = lax.cond(
                shape_factor == 1,
                lambda _: first_order_weighting(charge, x, y, z, rho, dx, dy, dz, x_wind, y_wind, z_wind),
                lambda _: second_order_weighting(charge, x, y, z, rho, dx, dy, dz, x_wind, y_wind, z_wind),
                operand=None
            )
            # use the shape factor to determine which weighting function to use and update rho accordingly
            return rho
        # add the particle species to the charge density array
        rho = jax.lax.fori_loop(0, N_particles, add_to_rho, rho)
    # The above loop iterates over each species of particles and updates the charge density array rho

    return rho

@jit
def first_order_weighting(q, x, y, z, rho, dx, dy, dz, x_wind, y_wind, z_wind):
    """
    Distribute the charge of a particle to the surrounding grid points.

    Args:
        q (float): Charge of the particle.
        x (float): x-coordinate of the particle.
        y (float): y-coordinate of the particle.
        z (float): z-coordinate of the particle.
        rho (ndarray): Charge density array.
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.
        x_wind (float): Window in the x-direction.
        y_wind (float): Window in the y-direction.
        z_wind (float): Window in the z-direction.

    Returns:
        ndarray: Updated charge density array.
    """

    Nx, Ny, Nz = rho.shape
    # get the shape of the charge density array

    x0 = jnp.floor((x + x_wind / 2) / dx).astype(int)
    y0 = jnp.floor((y + y_wind / 2) / dy).astype(int)
    z0 = jnp.floor((z + z_wind / 2) / dz).astype(int)
    # Calculate the nearest grid points

    deltax = x - jnp.floor(x / dx) * dx
    deltay = y - jnp.floor(y / dy) * dy
    deltaz = z - jnp.floor(z / dz) * dz
    # Calculate the difference between the particle position and the nearest grid point

    x1 = wrap_around(x0 + 1, Nx)
    y1 = wrap_around(y0 + 1, Ny)
    z1 = wrap_around(z0 + 1, Nz)
    # Calculate the index of the next grid point

    dv = dx * dy * dz
    # Calculate the volume of each grid point

    ##################################### WEIGHTING FACTORS ##########################################################
    Sx0 = 1 - (deltax/dx)
    Sy0 = 1 - (deltay/dy)
    Sz0 = 1 - (deltaz/dz)
    # calculate the weight for the center grid points
    Sx1 = deltax/dx
    Sy1 = deltay/dy
    Sz1 = deltaz/dz
    # calculate the weight for the next grid points
    ####################################################################################################################

    ####################################### CHARGE DISTRIBUTION ########################################################
    rho = rho.at[x0, y0, z0].add((q / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    rho = rho.at[x1, y0, z0].add((q / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    rho = rho.at[x0, y1, z0].add((q / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    rho = rho.at[x0, y0, z1].add((q / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    rho = rho.at[x1, y1, z0].add((q / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    rho = rho.at[x1, y0, z1].add((q / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    rho = rho.at[x0, y1, z1].add((q / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    rho = rho.at[x1, y1, z1].add((q / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    # Distribute the charge of the particle to the grid points
    #####################################################################################################################

    return rho

@jit
def second_order_weighting(q, x, y, z, rho, dx, dy, dz, x_wind, y_wind, z_wind):
    """
    Distribute the charge of a particle to the surrounding grid points using second-order weighting.

    Args:
        q (float): Charge of the particle.
        x (float): x-coordinate of the particle.
        y (float): y-coordinate of the particle.
        z (float): z-coordinate of the particle.
        rho (ndarray): Charge density array.
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.
        x_wind (float): Window in the x-direction.
        y_wind (float): Window in the y-direction.
        z_wind (float): Window in the z-direction.

    Returns:
        ndarray: Updated charge density array.
    """

    Nx, Ny, Nz = rho.shape
    # get the shape of the charge density array

    x0 = jnp.floor((x + x_wind / 2) / dx).astype(int)
    y0 = jnp.floor((y + y_wind / 2) / dy).astype(int)
    z0 = jnp.floor((z + z_wind / 2) / dz).astype(int)
    # Calculate the nearest grid points

    deltax = x - jnp.floor(x / dx) * dx
    deltay = y - jnp.floor(y / dy) * dy
    deltaz = z - jnp.floor(z / dz) * dz
    # Calculate the difference between the particle position and the nearest grid point

    x1 = wrap_around(x0 + 1, Nx)
    y1 = wrap_around(y0 + 1, Ny)
    z1 = wrap_around(z0 + 1, Nz)
    # Calculate the index of the next grid point

    x_minus1 = x0 - 1
    y_minus1 = y0 - 1
    z_minus1 = z0 - 1
    # Calculate the index of the previous grid point

    dv = dx * dy * dz
    # Calculate the volume of each grid point


    ####################### WEIGHTING FACTORS ############################################################################
    Sx0 = (3/4) - (deltax/dx)**2
    Sy0 = (3/4) - (deltay/dy)**2
    Sz0 = (3/4) - (deltaz/dz)**2
    # Calculate the weights for the central grid points

    Sx1 = jax.lax.cond(
        deltax <= dx/2,
        lambda _: (1/2) * ((1/2) - (deltax/dx))**2,
        lambda _: jnp.array(0.0, dtype=deltax.dtype),
        operand=None
    )
    Sy1 = jax.lax.cond(
        deltay <= dy/2,
        lambda _: (1/2) * ((1/2) - (deltay/dy))**2,
        lambda _: jnp.array(0.0, dtype=deltay.dtype),
        operand=None
    )
    Sz1 = jax.lax.cond(
        deltaz <= dz/2,
        lambda _: (1/2) * ((1/2) - (deltaz/dz))**2,
        lambda _: jnp.array(0.0, dtype=deltaz.dtype),
        operand=None
    )
    # Calculate the weights for the next grid points

    Sx_minus1 = jax.lax.cond(
        deltax <= dx/2,
        lambda _: (1/2) * ((1/2) + (deltax/dx))**2,
        lambda _: jnp.array(0.0, dtype=deltax.dtype),
        operand=None
    )

    Sy_minus1 = jax.lax.cond(
        deltay <= dy/2,
        lambda _: (1/2) * ((1/2) + (deltay/dy))**2,
        lambda _: jnp.array(0.0, dtype=deltay.dtype),
        operand=None
    )
    Sz_minus1 = jax.lax.cond(
        deltaz <= dz/2,
        lambda _: (1/2) * ((1/2) + (deltaz/dz))**2,
        lambda _: jnp.array(0.0, dtype=deltaz.dtype),
        operand=None
    )
    # Calculate the weights for the previous grid points
    #####################################################################################################################

    ###################################### CHARGE DISTRIBUTION ##########################################################
    rho = rho.at[x0, y0, z0].add((q / dv) * Sx0 * Sy0 * Sz0, mode='drop')
    rho = rho.at[x1, y0, z0].add((q / dv) * Sx1 * Sy0 * Sz0, mode='drop')
    rho = rho.at[x0, y1, z0].add((q / dv) * Sx0 * Sy1 * Sz0, mode='drop')
    rho = rho.at[x0, y0, z1].add((q / dv) * Sx0 * Sy0 * Sz1, mode='drop')
    rho = rho.at[x1, y1, z0].add((q / dv) * Sx1 * Sy1 * Sz0, mode='drop')
    rho = rho.at[x1, y0, z1].add((q / dv) * Sx1 * Sy0 * Sz1, mode='drop')
    rho = rho.at[x0, y1, z1].add((q / dv) * Sx0 * Sy1 * Sz1, mode='drop')
    rho = rho.at[x1, y1, z1].add((q / dv) * Sx1 * Sy1 * Sz1, mode='drop')
    rho = rho.at[x_minus1, y0, z0].add((q / dv) * Sx_minus1 * Sy0 * Sz0, mode='drop')
    rho = rho.at[x0, y_minus1, z0].add((q / dv) * Sx0 * Sy_minus1 * Sz0, mode='drop')
    rho = rho.at[x0, y0, z_minus1].add((q / dv) * Sx0 * Sy0 * Sz_minus1, mode='drop')
    rho = rho.at[x_minus1, y_minus1, z0].add((q / dv) * Sx_minus1 * Sy_minus1 * Sz0, mode='drop')
    rho = rho.at[x_minus1, y0, z_minus1].add((q / dv) * Sx_minus1 * Sy0 * Sz_minus1, mode='drop')
    rho = rho.at[x0, y_minus1, z_minus1].add((q / dv) * Sx0 * Sy_minus1 * Sz_minus1, mode='drop')
    rho = rho.at[x_minus1, y_minus1, z_minus1].add((q / dv) * Sx_minus1 * Sy_minus1 * Sz_minus1, mode='drop')
    # Distribute the charge of the particle to the grid points
    #####################################################################################################################


    return rho


@jit
def wrap_around(ix, size):
    """Wrap around index to ensure it is within bounds."""
    return jax.lax.cond(
        ix > size - 1,
        lambda _: ix - size,
        lambda _: ix,
        operand=None
    )


import jax
import jax.numpy as jnp
from jax import jit
from jax import lax
import functools
from functools import partial
#from memory_profiler import profile
# import external libraries

from PyPIC3D.pstd import spectral_poisson_solve, spectral_gradient
from PyPIC3D.fdtd import centered_finite_difference_gradient, solve_poisson_sor
from PyPIC3D.rho import compute_rho
#from PyPIC3D.sor import solve_poisson_sor
# import internal libraries

def initialize_fields(Nx, Ny, Nz):
    """
    Initializes the electric and magnetic field arrays, as well as the electric potential and charge density arrays.

    Args:
        Nx (int): Number of grid points in the x-direction.
        Ny (int): Number of grid points in the y-direction.
        Nz (int): Number of grid points in the z-direction.

    Returns:
        Ex (ndarray): Electric field array in the x-direction.
        Ey (ndarray): Electric field array in the y-direction.
        Ez (ndarray): Electric field array in the z-direction.
        Bx (ndarray): Magnetic field array in the x-direction.
        By (ndarray): Magnetic field array in the y-direction.
        Bz (ndarray): Magnetic field array in the z-direction.
        phi (ndarray): Electric potential array.
        rho (ndarray): Charge density array.
    """
    # get the number of grid points in each direction
    Ex = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    Ey = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    Ez = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    # initialize the electric field arrays as 0
    Bx = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    By = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    Bz = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    # initialize the magnetic field arrays as 0

    Jx = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    Jy = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    Jz = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    # initialize the current density arrays as 0

    phi = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    rho = jax.numpy.zeros(shape = (Nx, Ny, Nz) )
    # initialize the electric potential and charge density arrays as 0

    return (Ex, Ey, Ez), (Bx, By, Bz), (Jx, Jy, Jz), phi, rho

@partial(jit, static_argnums=(4, 5))
def solve_poisson(rho, constants, world, phi, solver, bc='periodic'):
    """
    Solve the Poisson equation for electrostatic potential.

    Args:
        rho (ndarray): Charge density.
        eps (float): Permittivity.
        dx (float): Grid spacing in the x-direction.
        dy (float): Grid spacing in the y-direction.
        dz (float): Grid spacing in the z-direction.
        phi (ndarray): Initial guess for the electrostatic potential.
        bc (str): Boundary condition.
        M (ndarray, optional): Preconditioner matrix for the conjugate gradient solver.

    Returns:
        phi (ndarray): Solution to the Poisson equation.
    """

    # phi = lax.cond(

    #     solver == 'spectral',

    #     lambda _: spectral_poisson_solve(rho, constants, world),
    #     lambda _: solve_poisson_sor(
    #         phi=phi,
    #         rho=rho,
    #         dx=world['dx'],
    #         dy=world['dy'],
    #         dz=world['dz'],
    #         eps=constants['eps'],
    #         omega=0.15,
    #         tol=1e-12,
    #         max_iter=15000
    #     ),

    #     operand=None
    # )

    phi = spectral_poisson_solve(rho, constants, world)

    # if solver == 'spectral':
    #     phi = spectral_poisson_solve()
    # elif solver == 'fdtd':
    #     phi = solve_poisson_sor()

    return phi

@partial(jit, static_argnums=(5, 6))
def calculateE(world, particles, constants, rho, phi, solver, bc):
    """
    Calculate the electric field components (Ex, Ey, Ez) and electric potential (phi)
    based on the given parameters.

    Args:
        E (tuple): Tuple containing the electric field components (Ex, Ey, Ez).
        world (dict): Dictionary containing the simulation world parameters such as
                    grid spacing (dx, dy, dz) and window dimensions (x_wind, y_wind, z_wind).
        particles (array): Array containing particle positions and properties.
        constants (dict): Dictionary containing physical constants such as permittivity (eps).
        rho (array): Charge density array.
        phi (array): Electric potential array.
        solver (str): Type of solver to use ('spectral' or other).
        bc (str): Boundary condition type.
        verbose (bool): Flag to enable verbose output.

    Returns:
        tuple: Updated electric field components (Ex, Ey, Ez), electric potential (phi), 
            and charge density (rho).
    """

    dx = world['dx']
    dy = world['dy']
    dz = world['dz']
    # get resolution

    rho = compute_rho(particles, rho, world)
    # calculate the charge density based on the particle positions

    phi = solve_poisson(rho=rho, constants=constants, world=world, phi=phi, solver=solver, bc=bc)
    # solve the Poisson equation to get the electric potential

    phi = digital_filter(phi, 0.5)
    # apply a digital filter to the electric potential
    # alpha = 0.5 is a bilinear filter for the electric potential

    Ex, Ey, Ez = lax.cond(
        solver == 'spectral',
        lambda _: spectral_gradient(-1*phi, world),
        lambda _: centered_finite_difference_gradient(-1*phi, dx, dy, dz, bc),
        operand=None
    )
    # compute the gradient of the electric potential to get the electric field

    return (Ex, Ey, Ez), phi, rho


@partial(jit, static_argnums=(5))
def update_E(E, B, J, world, constants, curl_func):
    """
    Update the electric field components (Ex, Ey, Ez) based on the given parameters.

    Args:
        grid (object): The grid object containing the simulation grid.
        staggered_grid (object): The staggered grid object for the simulation.
        E (tuple): A tuple containing the electric field components (Ex, Ey, Ez).
        B (tuple): A tuple containing the magnetic field components (Bx, By, Bz).
        J (tuple): A tuple containing the current density components (Jx, Jy, Jz).
        world (dict): A dictionary containing the world parameters such as 'dx', 'dy', 'dz', and 'dt'.
        constants (dict): A dictionary containing the physical constants such as 'C' (speed of light) and 'eps' (permittivity).
        curl_func (function): A function to calculate the curl of the magnetic field.

    Returns:
        tuple: Updated electric field components (Ex, Ey, Ez).
    """

    Ex, Ey, Ez = E
    Bx, By, Bz = B
    Jx, Jy, Jz = J
    # unpack the E, B, and J fields

    dt = world['dt']
    C = constants['C']
    eps = constants['eps']
    # get the time resolution and necessary constants

    curlx, curly, curlz = curl_func(Bx, By, Bz)
    # calculate the curl of the magnetic field

    Ex = Ex + ( C**2 * curlx - Jx / eps ) * dt
    Ey = Ey + ( C**2 * curly - Jy / eps ) * dt
    Ez = Ez + ( C**2 * curlz - Jz / eps ) * dt
    # update the electric field from Maxwell's equations

    return (Ex, Ey, Ez)


@partial(jit, static_argnums=(4))
def update_B(E, B, world, constants, curl_func):
    """
    Update the magnetic field components (Bx, By, Bz) using the curl of the electric field.

    Args:
        grid (ndarray): The grid on which the fields are defined.
        staggered_grid (ndarray): The staggered grid for field calculations.
        E (tuple): The electric field components (Ex, Ey, Ez).
        B (tuple): The magnetic field components (Bx, By, Bz).
        world (dict): Dictionary containing simulation parameters such as 'dx', 'dy', 'dz', and 'dt'.
        constants (dict): Dictionary containing physical constants.
        curl_func (function): Function to calculate the curl of the electric field.

    Returns:
        tuple: Updated magnetic field components (Bx, By, Bz).
    """

    dt = world['dt']
    # get the time resolution

    Ex, Ey, Ez = E
    Bx, By, Bz = B
    # unpack the E and B fields

    curlx, curly, curlz = curl_func(Ex, Ey, Ez)
    # calculate the curl of the electric field

    Bx = Bx - dt*curlx
    By = By - dt*curly
    Bz = Bz - dt*curlz
    # update the magnetic field from Maxwell's equations

    return (Bx, By, Bz)

@partial(jit, static_argnums=(1))
def digital_filter(phi, alpha):
    """
    Apply a digital filter to the electric potential array.

    Args:
        phi (ndarray): Electric potential array.
        alpha (float): Filter coefficient.

    Returns:
        ndarray: Filtered electric potential array.
    """
    filter_phi = alpha * phi +  (  (1 - alpha) / 6 ) * (  jnp.roll(phi, shift=1, axis=0) + jnp.roll(phi, shift=-1, axis=0) + \
                                                            jnp.roll(phi, shift=1, axis=1) + jnp.roll(phi, shift=-1, axis=1) + \
                                                                jnp.roll(phi, shift=1, axis=2) + jnp.roll(phi, shift=-1, axis=2) )
    return filter_phi


import jax.numpy as jnp
import jax
from jax import jit
from jax import lax
# import external libraries

from PyPIC3D.boundaryconditions import apply_zero_boundary_condition

def solve_poisson_sor(phi, rho, dx, dy, dz, eps, omega=1.5, tol=1e-6, max_iter=10000):
    """
    Solve Poisson's equation using Successive Over-Relaxation (SOR) method.

    Args:
        phi (jax.numpy.ndarray): Initial guess for the potential.
        rho (jax.numpy.ndarray): Charge density.
        omega (float): Relaxation factor.
        tol (float): Tolerance for convergence.
        max_iter (int): Maximum number of iterations.

    Returns:
        jax.numpy.ndarray: Solution for the potential.
    """
    phi = jnp.array(phi)
    rho = jnp.array(rho)
    b = dx * dy * dz / jnp.pi

    def cond_fun(val):
        _, norm_diff, iter_count = val
        return (norm_diff >= tol) & (iter_count < max_iter)

    def body_fun(val):
        phi, _, iter_count = val
        phi_old = phi.copy()
        phi = phi.at[1:-1, 1:-1, 1:-1].set(
            (1 - omega) * phi[1:-1, 1:-1, 1:-1] + omega / 6 * (
                phi[:-2, 1:-1, 1:-1] + phi[2:, 1:-1, 1:-1] +
                phi[1:-1, :-2, 1:-1] + phi[1:-1, 2:, 1:-1] +
                phi[1:-1, 1:-1, :-2] + phi[1:-1, 1:-1, 2:] -
                b * rho[1:-1, 1:-1, 1:-1] / eps
            )
        )
        return phi, jnp.linalg.norm(phi - phi_old), iter_count + 1

    phi, _, _ = jax.lax.while_loop(cond_fun, body_fun, (phi, jnp.inf, 0))

    return phi


#@partial(jit, static_argnums=(1, 2, 3, 4))
def centered_finite_difference_laplacian(field, dx, dy, dz, bc):
    """
    Calculates the Laplacian of a given field using centered finite difference and applies the specified boundary conditions.

    Args:
        field: numpy.ndarray
            The input field.
        dx: float
            The spacing between grid points in the x-direction.
        dy: float
            The spacing between grid points in the y-direction.
        dz: float
            The spacing between grid points in the z-direction.
        boundary_condition: str
            The type of boundary condition ('periodic', 'neumann', 'dirichlet').

    Returns:
        numpy.ndarray
            The Laplacian of the field with the specified boundary conditions applied.
    """

    if bc == 'dirichlet':
        field = apply_zero_boundary_condition(field)
        # apply zero boundary condition at the edges of the field

    x_comp = (jnp.roll(field, shift=1, axis=0) + jnp.roll(field, shift=-1, axis=0) - 2*field) / (dx*dx)
    y_comp = (jnp.roll(field, shift=1, axis=1) + jnp.roll(field, shift=-1, axis=1) - 2*field) / (dy*dy)
    z_comp = (jnp.roll(field, shift=1, axis=2) + jnp.roll(field, shift=-1, axis=2) - 2*field) / (dz*dz)
    # calculate the Laplacian of the field using centered finite difference

    if bc == 'neumann':
        x_comp = apply_zero_boundary_condition(x_comp)
        y_comp = apply_zero_boundary_condition(y_comp)
        z_comp = apply_zero_boundary_condition(z_comp)

    return x_comp + y_comp + z_comp

#@partial(jit, static_argnums=(3, 4, 5, 6))
def centered_finite_difference_curl(field_x, field_y, field_z, dx, dy, dz, bc):
    """
    Computes the curl of a vector field using centered finite differencing and applies the specified boundary conditions.

    Args:
        field_x: numpy.ndarray
            The x-component of the vector field.
        field_y: numpy.ndarray
            The y-component of the vector field.
        field_z: numpy.ndarray
            The z-component of the vector field.
        dx: float
            The spacing between grid points in the x-direction.
        dy: float
            The spacing between grid points in the y-direction.
        dz: float
            The spacing between grid points in the z-direction.
        bc: str
            The type of boundary condition ('periodic', 'neumann', 'dirichlet').

    Returns:
        tuple of numpy.ndarray
            The curl components (curl_x, curl_y, curl_z) of the vector field with the specified boundary conditions applied.
    """
    if bc == 'dirichlet':
        field_x = apply_zero_boundary_condition(field_x)
        field_y = apply_zero_boundary_condition(field_y)
        field_z = apply_zero_boundary_condition(field_z)

    dFz_dy = (jnp.roll(field_z, shift=-1, axis=1) - jnp.roll(field_z, shift=1, axis=1)) / (2 * dy)
    dFy_dz = (jnp.roll(field_y, shift=-1, axis=2) - jnp.roll(field_y, shift=1, axis=2)) / (2 * dz)
    dFx_dz = (jnp.roll(field_x, shift=-1, axis=2) - jnp.roll(field_x, shift=1, axis=2)) / (2 * dz)
    dFz_dx = (jnp.roll(field_z, shift=-1, axis=0) - jnp.roll(field_z, shift=1, axis=0)) / (2 * dx)
    dFy_dx = (jnp.roll(field_y, shift=-1, axis=0) - jnp.roll(field_y, shift=1, axis=0)) / (2 * dx)
    dFx_dy = (jnp.roll(field_x, shift=-1, axis=1) - jnp.roll(field_x, shift=1, axis=1)) / (2 * dy)

    curl_x = dFz_dy - dFy_dz
    curl_y = dFx_dz - dFz_dx
    curl_z = dFy_dx - dFx_dy
    # calculate the curl of the field

    if bc == 'neumann':
        curl_x = apply_zero_boundary_condition(curl_x)
        curl_y = apply_zero_boundary_condition(curl_y)
        curl_z = apply_zero_boundary_condition(curl_z)

    return curl_x, curl_y, curl_z

#@partial(jit, static_argnums=(1, 2, 3, 4))
def centered_finite_difference_gradient(field, dx, dy, dz, bc):
    """
    Computes the gradient of a scalar field using centered finite differencing and applies the specified boundary conditions.

    Args:
        field: numpy.ndarray
            The input scalar field.
        dx: float
            The spacing between grid points in the x-direction.
        dy: float
            The spacing between grid points in the y-direction.
        dz: float
            The spacing between grid points in the z-direction.
        bc: str
            The type of boundary condition ('periodic', 'neumann', 'dirichlet').

    Returns:
        tuple of numpy.ndarray
            The gradient components (grad_x, grad_y, grad_z) of the scalar field with the specified boundary conditions applied.
    """
    if bc == 'dirichlet':
        field = apply_zero_boundary_condition(field)

    grad_x = (jnp.roll(field, shift=-1, axis=0) - jnp.roll(field, shift=1, axis=0)) / (2 * dx)
    grad_y = (jnp.roll(field, shift=-1, axis=1) - jnp.roll(field, shift=1, axis=1)) / (2 * dy)
    grad_z = (jnp.roll(field, shift=-1, axis=2) - jnp.roll(field, shift=1, axis=2)) / (2 * dz)

    if bc == 'neumann':
        grad_x = apply_zero_boundary_condition(grad_x)
        grad_y = apply_zero_boundary_condition(grad_y)
        grad_z = apply_zero_boundary_condition(grad_z)

    return grad_x, grad_y, grad_z

#@partial(jit, static_argnums=(3, 4, 5, 6))
def centered_finite_difference_divergence(field_x, field_y, field_z, dx, dy, dz, bc):
    """
    Computes the divergence of a vector field using centered finite differencing and applies the specified boundary conditions.

    Args:
        field_x: numpy.ndarray
            The x-component of the vector field.
        field_y: numpy.ndarray
            The y-component of the vector field.
        field_z: numpy.ndarray
            The z-component of the vector field.
        dx: float
            The spacing between grid points in the x-direction.
        dy: float
            The spacing between grid points in the y-direction.
        dz: float
            The spacing between grid points in the z-direction.
        bc: str
            The type of boundary condition ('periodic', 'neumann', 'dirichlet').

    Returns:
        numpy.ndarray
            The divergence of the vector field with the specified boundary conditions applied.
    """
    if bc == 'dirichlet':
        field_x = apply_zero_boundary_condition(field_x)
        field_y = apply_zero_boundary_condition(field_y)
        field_z = apply_zero_boundary_condition(field_z)

    div_x = (jnp.roll(field_x, shift=-1, axis=0) - jnp.roll(field_x, shift=1, axis=0)) / (2 * dx)
    div_y = (jnp.roll(field_y, shift=-1, axis=1) - jnp.roll(field_y, shift=1, axis=1)) / (2 * dy)
    div_z = (jnp.roll(field_z, shift=-1, axis=2) - jnp.roll(field_z, shift=1, axis=2)) / (2 * dz)

    if bc == 'neumann':
        div_x = apply_zero_boundary_condition(div_x)
        div_y = apply_zero_boundary_condition(div_y)
        div_z = apply_zero_boundary_condition(div_z)

    return div_x + div_y + div_z